# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

__all__ = [
    "BasePOD",
    "BoolPOD",
    "DatetimePOD",
    "EnumPOD",
    "FloatPOD",
    "HTMLStringPOD",
    "IntPOD",
    "MultiStringPOD",
    "StringPOD",
]

import abc
import collections.abc as cabc
import datetime
import enum
import itertools
import math
import os
import re
import typing as t

import markupsafe
import typing_extensions as te
from lxml import etree

from capellambse import helpers

from . import E, U

if t.TYPE_CHECKING:
    from . import _obj


class BasePOD(t.Generic[U]):
    """A plain-old-data descriptor."""

    __slots__ = (
        "__dict__",
        "__name__",
        "__objclass__",
        "attribute",
        "default",
        "writable",
    )

    NOT_OPTIONAL = object()

    def __init__(
        self,
        attribute: str,
        *,
        default: U,
        writable: bool,
    ) -> None:
        self.attribute = attribute
        self.writable = writable
        self.default = default

        self.__name__ = "(unknown)"
        self.__objclass__: type[t.Any] | None = None

    @property
    def _qualname(self) -> str:
        """Generate the qualified name of this descriptor."""
        if self.__objclass__ is None:
            return f"(unknown {type(self).__name__} - call __set_name__)"
        return f"{self.__objclass__.__name__}.{self.__name__}"

    @t.overload
    def __get__(self, obj: None, objtype: type) -> te.Self: ...
    @t.overload
    def __get__(self, obj: t.Any, objtype: type | None = None) -> U: ...
    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:
            return self

        data = obj._element.get(self.attribute)
        assert data is None or isinstance(data, str)
        if data is None:
            return self.default
        return self._from_xml(obj, data)

    def __set__(self, obj: t.Any, value: U | None) -> None:
        if not self.writable and self.attribute in obj._element.attrib:
            raise TypeError(f"{self._qualname} is not writable")

        if value is not None and value != self.default:
            data = self._to_xml(obj, value)
        else:
            data = None

        if data is None:
            obj._element.attrib.pop(self.attribute, None)
        else:
            obj._element.attrib[self.attribute] = data

    def __delete__(self, obj: t.Any) -> None:
        self.__set__(obj, None)

    def __set_name__(self, owner: type[t.Any], name: str) -> None:
        if self.__objclass__ is not None:
            raise RuntimeError(
                f"__set_name__ called twice on {self._qualname}"
            )
        self.__name__ = name
        self.__objclass__ = owner

    @abc.abstractmethod
    def _from_xml(self, obj: _obj.ModelElement, value: str, /) -> U: ...
    @abc.abstractmethod
    def _to_xml(self, obj: _obj.ModelElement, value: U, /) -> str | None: ...


class StringPOD(BasePOD[str]):
    """A POD containing arbitrary string data."""

    __slots__ = ()

    def __init__(self, attribute: str, /, *, writable: bool = True) -> None:
        """Create a StringPOD.

        Parameters
        ----------
        attribute
            The name of the attribute on the XML element.
        writable
            Whether to allow changing the value at runtime.
        """
        super().__init__(attribute, default="", writable=writable)

    def _from_xml(self, obj: _obj.ModelElement, value: str, /) -> str:
        del obj
        return value

    def _to_xml(self, obj: _obj.ModelElement, value: str, /) -> str:
        del obj
        return value


class HTMLStringPOD(BasePOD[markupsafe.Markup]):
    """A POD containing a string with HTML inside."""

    __slots__ = ()

    def __init__(self, attribute: str, /, *, writable: bool = True) -> None:
        """Create an HTMLStringPOD.

        Parameters
        ----------
        attribute
            The name of the attribute on the XML element.
        writable
            Whether to allow changing the value at runtime.
        """
        super().__init__(
            attribute,
            default=markupsafe.Markup(""),
            writable=writable,
        )

    def _from_xml(
        self, obj: _obj.ModelElement, value: str, /
    ) -> markupsafe.Markup:
        if os.getenv("CAPELLAMBSE_XHTML") == "1":
            value = helpers.repair_html(value)
        value = helpers.embed_images(value, obj._model.resources)
        return markupsafe.Markup(value)

    def _to_xml(
        self, obj: _obj.ModelElement, value: markupsafe.Markup, /
    ) -> str:
        value = helpers.repair_html(value)
        return helpers.unembed_images(value, obj._model.resources)


class BoolPOD(BasePOD[bool]):
    """A POD containing a boolean."""

    __slots__ = ()

    def __init__(self, attribute: str, /) -> None:
        """Create a BoolPOD.

        Parameters
        ----------
        attribute
            The name of the attribute on the XML element.
        """
        super().__init__(attribute, default=False, writable=True)

    def _from_xml(self, obj: _obj.ModelElement, value: str, /) -> bool:
        del obj
        return value == "true"

    def _to_xml(self, obj: _obj.ModelElement, value: bool, /) -> str:
        del obj
        assert isinstance(value, bool)
        return ("false", "true")[value]


class IntPOD(BasePOD[int]):
    """A POD containing an integer number."""

    __slots__ = ()

    def __init__(self, attribute: str, /, *, writable: bool = True) -> None:
        """Create an IntPOD.

        Parameters
        ----------
        attribute
            The name of the attribute on the XML element.
        writable
            Whether to allow changing the value at runtime.
        """
        super().__init__(attribute, default=0, writable=writable)

    def _from_xml(self, obj: _obj.ModelElement, data: str, /) -> int:
        del obj
        return int(data)

    def _to_xml(self, obj: _obj.ModelElement, value: int, /) -> str | None:
        del obj
        if not isinstance(value, int):
            raise TypeError(
                f"{self._qualname} only accepts int,"
                f" not {type(value).__name__}"
            )
        return str(value)


class FloatPOD(BasePOD[float]):
    """A POD containing a floating-point number.

    In Capella's Java land, these are often called "real numbers".
    """

    __slots__ = ()

    def __init__(self, attribute: str, /, *, writable: bool = True) -> None:
        """Create an IntPOD.

        Parameters
        ----------
        attribute
            The name of the attribute on the XML element.
        writable
            Whether to allow changing the value at runtime.
        """
        super().__init__(attribute, default=0.0, writable=writable)

    def _from_xml(self, obj: _obj.ModelElement, data: str, /) -> float:
        del obj
        return float(data)

    def _to_xml(self, obj: _obj.ModelElement, value: float, /) -> str | None:
        del obj
        if isinstance(value, int):
            value = float(value)
        elif not isinstance(value, float):
            raise TypeError(
                f"{self._qualname} only accepts float or int,"
                f" not {type(value).__name__}"
            )
        assert isinstance(value, float)

        if math.isnan(value):
            raise ValueError("Cannot represent NaN")
        if value == math.inf:
            return "*"
        if value == -math.inf:
            raise ValueError("Cannot represent negative infinity")
        return str(value)


class DatetimePOD(BasePOD[datetime.datetime | None]):
    """A POD containing a timestamp.

    The value stored in the XML will be formatted as required by
    Capella. This format is the ISO8601 format with millisecond
    precision, but no ``:`` in the time zone specification.
    """

    __slots__ = ()
    re_set = re.compile(r"(?<=[+-]\d\d):(?=\d\d$)")
    re_get = re.compile(r"(?<=[+-]\d\d)(?=\d\d$)")

    def __init__(self, attribute: str, /, *, writable: bool = True) -> None:
        """Create a DatetimePOD.

        Parameters
        ----------
        attribute
            The name of the attribute on the XML element.
        writable
            Whether to allow changing the value at runtime.
        """
        super().__init__(attribute, default=None, writable=writable)

    def _from_xml(
        self, obj: _obj.ModelElement, value: str, /
    ) -> datetime.datetime:
        del obj
        formatted = self.re_get.sub(":", value)
        return datetime.datetime.fromisoformat(formatted)

    def _to_xml(
        self, obj: _obj.ModelElement, value: datetime.datetime | None, /
    ) -> str:
        del obj
        assert value is not None
        if not isinstance(value, datetime.datetime):
            raise TypeError(f"Expected datetime instance, not {value!r}")
        if value.tzinfo is None or value.tzinfo.utcoffset(value) is None:
            value = value.astimezone()

        formatted = value.isoformat("T", "milliseconds")
        return self.re_set.sub("", formatted)


class EnumPOD(BasePOD[E]):
    """A POD that can have one of a pretermined set of values.

    This works in much the same way as the StringPOD, except that the
    returned and consumed values are not simple strings, but members of
    the Enum that was passed into the constructor.

    When assigning, this property also accepts the string name of one of
    the enum members. In either case, the corresponding enum member's
    value will be placed in the underlying XML element.

    The *default* constructor argument determines which member will be
    used if the attribute is missing from the XML. If no *default* is
    passed exlicitly, the first declared enum member will be used.
    """

    __slots__ = ("enumcls",)

    def __init__(
        self,
        attribute: str,
        enumcls: type[E],
        /,
        default: E | str | None = None,
        *,
        writable: bool = True,
    ) -> None:
        """Create an EnumPOD.

        Parameters
        ----------
        attribute
            The name of the attribute on the XML element.
        enumcls
            The :class:`enum.Enum` subclass to use. The class' members'
            values are used as the possible values for the XML
            attribute.
        default
            The default value to return if the attribute is not present
            in the XML. If not passed or None, the first member of the
            given *enumcls* will be used.
        writable
            Whether to allow changing the value at runtime.
        """
        if not (isinstance(enumcls, type) and issubclass(enumcls, enum.Enum)):
            raise TypeError(
                f"enumcls must be an Enum subclass, not {enumcls!r}"
            )

        if default is None:
            try:
                default = next(iter(enumcls.__members__.values()))
            except StopIteration:
                raise TypeError(
                    f"Enum class does not have any members: {enumcls!r}"
                ) from None
        elif isinstance(default, str):
            default = enumcls[default]
        elif not isinstance(default, enumcls):
            raise TypeError(
                f"'default' must be a member of 'enumcls', not {default!r}"
            )
        assert isinstance(default, enumcls)

        super().__init__(attribute, default=default, writable=writable)
        self.enumcls = enumcls

    def _from_xml(self, obj: _obj.ModelElement, value: str, /) -> E:
        del obj
        return self.enumcls(value)

    def _to_xml(self, obj: _obj.ModelElement, value: E | str, /) -> str | None:
        del obj
        if isinstance(value, str):
            value = self.enumcls[value]
        return value.value


class MultiStringPOD(BasePOD[cabc.MutableSequence[str]]):
    """A POD that provides access to a list of strings."""

    __slots__ = (
        "__name__",
        "__objclass__",
        "tag",
    )

    def __init__(self, tag: str) -> None:
        super().__init__(tag, default=[], writable=True)
        self.tag = tag

        self.__name__ = "(unknown)"
        self.__objclass__: type[t.Any] | None = None

    @t.overload
    def __get__(self, obj: None, objtype: type) -> te.Self: ...
    @t.overload
    def __get__(
        self, obj: t.Any, objtype: type | None = None
    ) -> cabc.MutableSequence[str]: ...
    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:
            return self

        return _MultiPODValues(obj._element, self.tag)

    def __set__(self, obj: t.Any, value: cabc.Sequence[str] | None) -> None:
        if value is None:
            value = ()
        self.__get__(obj)[:] = value

    def __delete__(self, obj: t.Any) -> None:
        self.__get__(obj)[:] = ()

    def _from_xml(
        self, obj: _obj.ModelElement, value: str, /
    ) -> cabc.MutableSequence[str]:
        del obj, value
        raise TypeError("Not used for this POD type")

    def _to_xml(
        self, obj: _obj.ModelElement, value: cabc.MutableSequence[str], /
    ) -> str | None:
        del obj, value
        raise TypeError("Not used for this POD type")


class _MultiPODValues(cabc.MutableSequence[str]):
    def __init__(self, parent: etree._Element, tag: str) -> None:
        self._parent = parent
        self._tag = tag

    @t.overload
    def __getitem__(self, _: int, /) -> str: ...
    @t.overload
    def __getitem__(self, _: slice, /) -> cabc.MutableSequence[str]: ...
    def __getitem__(
        self, i: int | slice, /
    ) -> str | cabc.MutableSequence[str]:
        values = list(self._parent.iterchildren(self._tag))
        if isinstance(i, slice):
            return [v.text or "" for v in values[i]]
        return values[i].text or ""

    @t.overload
    def __setitem__(self, i: int, v: str, /) -> None: ...
    @t.overload
    def __setitem__(self, i: slice, v: cabc.Iterable[str], /) -> None: ...
    def __setitem__(
        self, idx: int | slice, value: str | cabc.Iterable[str], /
    ) -> None:
        if not isinstance(idx, int | slice):
            raise TypeError(
                "indices must be integers or slices, not {type(i).__name__}"
            )

        if isinstance(idx, slice) and isinstance(value, cabc.Iterable):
            it = self._parent.iterchildren(self._tag)
            elems = list(itertools.islice(it, idx.start, idx.stop, idx.step))
            values = [i.text for i in elems]
            values[idx] = value
            for e, v in zip(elems, values, strict=True):
                e.text = v

        elif isinstance(idx, int) and isinstance(value, str):
            if idx < 0:
                idx += len(self)
                if idx < 0:
                    raise IndexError("assignment index out of range")

            it = self._parent.iterchildren(self._tag)
            elem = next(itertools.islice(it, idx, idx + 1), None)
            if elem is None:
                raise IndexError("assignment index out of range") from None
            elem.text = value

    def __delitem__(self, i: int | slice, /) -> None:
        it = self._parent.iterchildren(self._tag)
        if isinstance(i, slice):
            elems = list(itertools.islice(it, i.start, i.stop, i.step))
        else:
            elems = list(itertools.islice(it, i, i + 1, 1))
        for e in elems:
            self._parent.remove(e)

    def __len__(self) -> int:
        return sum(1 for _ in self._parent.iterchildren(self._tag))

    def insert(self, index: int, value: str, /) -> None:
        if index < 0:
            index = max(index + len(self), 0)
        e: etree._Element | None = None

        container = self._parent.makeelement(self._tag)
        container.text = value

        for i, e in enumerate(self._parent.iterchildren(self._tag)):
            if i == index:
                e.addprevious(container)
                break
        else:
            if e is not None:
                e.addnext(container)
                e.append(container)



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""An Eclipse-like XML serializer.

The libxml2 XML serializer produces very different output from the one
used by Capella. This causes a file saved by libxml2 to look vastly
different, even though semantically nothing might have changed at all.
This module implements a serializer which produces output like Capella
does.
"""

from __future__ import annotations

import collections.abc as cabc
import contextlib
import html.entities
import io
import os
import re
import sys
import typing as t
import warnings

import lxml.etree

try:
    from capellambse._compiled import serialize as _native_serialize
except ImportError:
    if (
        os.environ.get("CIBUILDWHEEL", "0") == "1"
        or os.environ.get("CAPELLAMBSE_REQUIRE_NATIVE", "0") == "1"
    ):
        raise

    def _native_serialize(*_1, **_2):  # type: ignore[misc]
        raise TypeError("Native module is not available")

    HAS_NATIVE = False
else:
    HAS_NATIVE = True

_UnspecifiedType = t.NewType("_UnspecifiedType", object)
_NOT_SPECIFIED = _UnspecifiedType(object())

INDENT = b"  "
LINESEP = os.linesep.encode("ascii")
LINE_LENGTH = 80

ESCAPE_CHARS = r"[\x00-\x08\x0A-\x1F\x7F{}]"
P_ESCAPE_TEXT = re.compile(ESCAPE_CHARS.format('"&<'))
P_ESCAPE_COMMENTS = re.compile(ESCAPE_CHARS.format(">"))
P_ESCAPE_ATTR = re.compile(ESCAPE_CHARS.format('"&<\x09'))
P_NAME = re.compile(r"^(?:\{([^}]*)\})?(.+)$")

ALWAYS_EXPANDED_TAGS = frozenset({"bodies", "semanticResources"})


@t.runtime_checkable
class HasWrite(t.Protocol):
    """A simple protocol to check for a writable file-like object."""

    def write(self, chunk: bytes) -> t.Any: ...


def to_string(tree: lxml.etree._Element, /) -> str:
    """Serialize an XML tree as a ``str``.

    No XML processing instruction will be inserted at the start of the
    document.

    Arguments
    ---------
    tree
        The XML tree to serialize.

    Returns
    -------
    str
        The serialized XML.
    """
    payload = serialize(tree, encoding="utf-8", errors="surrogateescape")
    return payload.decode("utf-8", errors="surrogateescape")


def to_bytes(
    tree: lxml.etree._Element,
    /,
    *,
    encoding: str | _UnspecifiedType = _NOT_SPECIFIED,
    errors: str | _UnspecifiedType = _NOT_SPECIFIED,
    declare_encoding: bool = True,
) -> bytes:
    """Serialize an XML tree as a ``str``.

    At the start of the document, an XML processing instruction will be
    inserted declaring the used encoding. Pass
    ``declare_encoding=False`` to inhibit this behavior.

    Arguments
    ---------
    tree
        The XML tree to serialize.
    encoding
        The encoding to use. An XML processing instruction will be
        inserted which declares the used encoding.
    errors
        How to handle errors during encoding.

    Returns
    -------
    bytes
        The serialized XML, encoded using ``encoding``.
    """
    args = {}
    if encoding is not _NOT_SPECIFIED:
        assert isinstance(encoding, str)
        args["encoding"] = encoding
    else:
        encoding = "utf-8"
    if errors is not _NOT_SPECIFIED:
        assert isinstance(errors, str)
        args["errors"] = errors

    if declare_encoding:
        declaration = _declare(encoding)
    else:
        declaration = b""
    return declaration + serialize(tree, **args)  # type: ignore[call-overload]


def write(
    tree: lxml.etree._Element | lxml.etree._ElementTree,
    /,
    file: HasWrite | os.PathLike | str | bytes,
    *,
    encoding: str | _UnspecifiedType = _NOT_SPECIFIED,
    errors: str | _UnspecifiedType = _NOT_SPECIFIED,
    line_length: float | int = LINE_LENGTH,
    siblings: bool = False,
) -> None:
    """Write the XML tree to ``file``.

    Parameters
    ----------
    tree
        The XML tree to serialize.
    file
        An open file or a PathLike to write the XML into.
    encoding
        The file encoding to use when opening a file.
    errors
        Set the encoding error handling behavior of newly opened files.
    line_length
        The number of characters after which to force a line break.
    siblings
        Also include siblings of the given subtree.
    """
    args = {}
    if encoding is not _NOT_SPECIFIED:
        assert isinstance(encoding, str)
        args["encoding"] = encoding
    else:
        encoding = "utf-8"
    if errors is not _NOT_SPECIFIED:
        assert isinstance(errors, str)
        args["errors"] = errors

    ctx: t.ContextManager[HasWrite]
    if isinstance(file, HasWrite):
        ctx = contextlib.nullcontext(file)
    else:
        ctx = open(file, "wb")  # noqa: SIM115

    with ctx as f:
        f.write(_declare(encoding))
        serialize(
            tree,
            **args,
            line_length=line_length,
            siblings=siblings,
            file=f,
        )


@t.overload
def serialize(
    tree: lxml.etree._Element | lxml.etree._ElementTree,
    /,
    *,
    encoding: str = ...,
    errors: str = ...,
    line_length: float | int = ...,
    siblings: bool | None = ...,
    file: None = ...,
) -> bytes: ...
@t.overload
def serialize(
    tree: lxml.etree._Element | lxml.etree._ElementTree,
    /,
    *,
    encoding: str = ...,
    errors: str = ...,
    line_length: float | int = ...,
    siblings: bool | None = ...,
    file: HasWrite,
) -> None: ...
def serialize(
    tree: lxml.etree._Element | lxml.etree._ElementTree,
    /,
    *,
    encoding: str | _UnspecifiedType = _NOT_SPECIFIED,
    errors: str | _UnspecifiedType = _NOT_SPECIFIED,
    line_length: float | int = LINE_LENGTH,
    siblings: bool | None = None,
    file: HasWrite | None = None,
) -> bytes | None:
    """Serialize an XML tree.

    The iterator returned by this function yields the serialized XML
    piece by piece.

    Parameters
    ----------
    tree
        The XML tree to serialize.
    encoding
        The encoding to use when generating XML.
    errors
        The encoding error handling behavior.
    line_length
        The number of characters after which to force a line break.
    siblings
        Also include siblings of the given subtree. Defaults to yes if
        'tree' is an element tree, no if it's a single element.
    file
        A file-like object to write the serialized tree to. If None, the
        serialized tree will be returned as bytes instead.

    Returns
    -------
    bytes | None
        The serialized tree (if no *file* was given), or None.
    """
    root: lxml.etree._Element
    if isinstance(tree, lxml.etree._ElementTree):
        if siblings is None:
            siblings = True
        root = tree.getroot()
    else:
        if siblings is None:
            siblings = False
        root = tree

    if encoding is not _NOT_SPECIFIED or errors is not _NOT_SPECIFIED:
        warnings.warn(
            (
                "The 'encoding' and 'errors' arguments are deprecated."
                " The default of strict UTF-8 will soon become the only option."
                " Please remove the arguments from your calls into exs."
            ),
            DeprecationWarning,
            stacklevel=2,
        )

    if encoding is _NOT_SPECIFIED:
        encoding = "utf-8"
    if errors is _NOT_SPECIFIED:
        errors = "strict"
    assert isinstance(encoding, str)
    assert isinstance(errors, str)

    if HAS_NATIVE and encoding == "utf-8" and errors == "strict":
        line_length = min(line_length, sys.maxsize)
        return _native_serialize(
            root, line_length=int(line_length), siblings=siblings, file=file
        )
    return _python_serialize(
        root,
        encoding=encoding,
        errors=errors,
        line_length=line_length,
        siblings=siblings,
        file=file,
    )


def _python_serialize(
    root,
    *,
    encoding: str,
    errors: str,
    line_length: float | int,
    siblings: bool,
    file: HasWrite | None,
) -> bytes | None:
    buffer = io.BytesIO()
    preceding_siblings: cabc.Iterable[lxml.etree._Element]
    following_siblings: cabc.Iterable[lxml.etree._Element]

    if siblings:
        preceding_siblings = reversed(list(root.itersiblings(preceding=True)))
        following_siblings = root.itersiblings()
    else:
        preceding_siblings = ()
        following_siblings = ()

    pos = 0
    for i in preceding_siblings:
        assert isinstance(i, lxml.etree._Comment), "Non-comment before tree"
        pos = _serialize_comment(
            buffer, i, encoding=encoding, errors=errors, pos=pos, indent=0
        )

    _serialize_element(
        buffer,
        root,
        0,
        encoding=encoding,
        errors=errors,
        line_length=line_length,
        pos=pos,
    )
    if (root.tail or "").strip():
        pos = _serialize_text(
            buffer,
            root.tail,
            encoding=encoding,
            errors=errors,
            pos=pos,
            multiline=True,
            escape_pattern=P_ESCAPE_TEXT,
        )

    for i in following_siblings:
        assert isinstance(i, lxml.etree._Comment), "Non-comment after tree"
        pos = _serialize_comment(
            buffer, i, encoding=encoding, errors=errors, pos=pos, indent=0
        )

    buffer.write(LINESEP)

    if file is not None:
        file.write(buffer.getvalue())
        return None
    return buffer.getvalue()


def _declare(encoding: str) -> bytes:
    return b"".join(
        (
            b'<?xml version="1.0" encoding="',
            encoding.upper().encode("ascii"),
            b'"?>',
            os.linesep.encode("ascii"),
        )
    )


def _escape(string: str, *, pattern: re.Pattern[str]) -> str:
    return pattern.sub(_escape_char, string)


def _escape_char(
    match: re.Match[str], *, ord_low: int = ord(" "), ord_high: int = ord("~")
) -> str:
    char = match.group(0)
    assert len(char) == 1
    if ord_low <= ord(char) <= ord_high:
        return f"&{html.entities.codepoint2name[ord(char)]};"
    return f"&#x{ord(char):X};"


def _unmapped_attrs(
    nsmap: cabc.Mapping[str, str], element: lxml.etree._Element
) -> cabc.Iterator[tuple[str, str]]:
    parent = element.getparent()
    if parent is None:
        parent_ns = set()
    else:
        parent_ns = set(parent.nsmap)

    attribs = dict(element.items())
    for attr in (
        "{http://www.omg.org/XMI}version",
        "{http://www.omg.org/XMI}type",
        "{http://www.omg.org/XMI}id",
        "{http://www.w3.org/2001/XMLSchema-instance}type",
    ):
        value = attribs.pop(attr, None)
        if value is not None:
            yield (
                _unmap_namespace(nsmap, attr),
                _escape(value, pattern=P_ESCAPE_ATTR),
            )

    assert None not in element.nsmap
    for nsname, value in sorted(element.nsmap.items(), key=_ns_sortkey):
        if nsname in parent_ns:
            continue
        yield (f"xmlns:{nsname}", value)

    for attr, value in attribs.items():
        yield (
            _unmap_namespace(nsmap, attr),
            _escape(value, pattern=P_ESCAPE_ATTR),
        )


def _ns_sortkey(v: tuple[str | None, str]) -> tuple[int, str | None]:
    ns, _ = v
    if ns == "xmi":
        return (0, ns)
    if ns == "xsi":
        return (1, ns)
    return (2, ns)


def _serialize_comment(
    buffer: HasWrite,
    comment: lxml.etree._Comment,
    /,
    *,
    encoding: str,
    errors: str,
    pos: int,
    indent: int,
) -> int:
    assert isinstance(comment, lxml.etree._Comment)

    buffer.write(LINESEP)
    buffer.write(INDENT * indent)
    buffer.write(b"<!--")
    pos = _serialize_text(
        buffer,
        comment.text,
        encoding=encoding,
        errors=errors,
        pos=len(INDENT) * indent,
        escape_pattern=P_ESCAPE_COMMENTS,
    )
    buffer.write(b"-->")

    if (comment.tail or "").strip():
        pos = _serialize_text(
            buffer,
            comment.tail,
            encoding=encoding,
            errors=errors,
            pos=pos,
            escape_pattern=P_ESCAPE_TEXT,
        )
    else:
        buffer.write(LINESEP)
        buffer.write(INDENT * indent)
        pos = len(INDENT) * indent
    return pos


def _serialize_element(
    buffer: HasWrite,
    element: lxml.etree._Element,
    indent: int,
    *,
    encoding: str,
    errors: str,
    pos: int = 0,
    line_length: float | int,
) -> int:
    assert isinstance(element, lxml.etree._Element)
    assert None not in element.nsmap
    nsmap = {v: k for k, v in element.nsmap.items() if k}
    buffer.write(b"<")
    assert isinstance(element.tag, str)
    tag = _unmap_namespace(nsmap, element.tag).encode(encoding, errors)
    buffer.write(tag)

    pos += 1 + len(tag)
    attr_indent = INDENT * (indent + 2)
    force_break = False
    for attr, value in _unmapped_attrs(nsmap, element):
        if pos > line_length or force_break:
            buffer.write(LINESEP)
            buffer.write(attr_indent)
            pos = len(attr_indent)
            force_break = False
        else:
            buffer.write(b" ")
            pos += 1

        buffer.write(attr.encode(encoding, errors))
        buffer.write(b'="')
        buffer.write(value.encode(encoding, errors))
        buffer.write(b'"')
        pos += len(attr) + len(value) + 3

        if element.getparent() is None and attr == "id":
            force_break = True

    if (
        element.text is None
        and len(element) == 0
        and element.tag not in ALWAYS_EXPANDED_TAGS
    ):
        buffer.write(b"/>")
        return pos + 2
    buffer.write(b">")

    child_indent = INDENT * (indent + 1)
    if (element.text or "").strip():
        pos = _serialize_text(
            buffer,
            element.text,
            encoding=encoding,
            errors=errors,
            pos=pos,
            multiline=True,
            escape_pattern=P_ESCAPE_TEXT,
        )
        text_content = True
    else:
        text_content = False

    for child in element:
        if not text_content:
            buffer.write(LINESEP)
            buffer.write(child_indent)
            pos = len(child_indent)

        pos = _serialize_element(
            buffer,
            child,
            indent + 1,
            encoding=encoding,
            errors=errors,
            pos=pos,
            line_length=line_length,
        )
        if (element.tail or "").strip():
            pos = _serialize_text(
                buffer,
                element.tail,
                encoding=encoding,
                errors=errors,
                pos=pos,
                escape_pattern=P_ESCAPE_TEXT,
            )
            text_content = True
        else:
            text_content = False

    if len(element) > 0 and not text_content:
        buffer.write(LINESEP)
        buffer.write(INDENT * indent)
        pos = len(INDENT) * indent

    buffer.write(b"</")
    buffer.write(tag)
    buffer.write(b">")
    return pos + len(tag) + 3


def _serialize_text(
    buffer: HasWrite,
    text: str | None,
    /,
    *,
    encoding: str,
    errors: str,
    pos: int,
    multiline: bool = False,
    escape_pattern: re.Pattern[str],
) -> int:
    if not text:
        return pos
    i, line = 0, ""
    for i, line in enumerate(text.split("\n")):
        if multiline and i:
            buffer.write(LINESEP)
        buffer.write(
            _escape(line, pattern=escape_pattern).encode(encoding, errors)
        )
    return len(line) + bool(i) * pos


def _unmap_namespace(nsmap: cabc.Mapping[str, str], name: str) -> str:
    match = P_NAME.search(name)
    assert match is not None
    if ns := match.group(1) or "":
        try:
            ns = nsmap[ns]
        except KeyError:
            raise ValueError(f"Namespace not found: {ns!r}") from None
        assert ns

    tag = match.group(2)
    assert tag

    return "".join((ns, ns and ":", tag))



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""CLI for the diagram cache updating feature."""

from __future__ import annotations

import subprocess
import sys
import warnings

from capellambse._diagram_cache import IndexEntry as IndexEntry


def _main() -> None:
    name = "export-diagrams"
    warnings.warn(
        f"This CLI entry point is deprecated, use 'capellambse {name}' instead",
        FutureWarning,
        stacklevel=2,
    )

    cmd = [sys.executable, "-mcapellambse", name, *sys.argv[1:]]
    raise SystemExit(subprocess.run(cmd, check=False).returncode)


if __name__ == "__main__":
    _main()



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import enum

import capellambse.model as m

from . import namespaces as ns

NS = ns.RE


@m.stringy_enum
@enum.unique
class CatalogElementKind(enum.Enum):
    REC = "REC"
    RPL = "RPL"
    REC_RPL = "REC_RPL"
    GROUPING = "GROUPING"


class ReAbstractElement(m.ModelElement, abstract=True):
    pass


class ReNamedElement(ReAbstractElement, abstract=True):
    name = m.StringPOD("name")


class ReDescriptionElement(ReNamedElement, abstract=True):
    description = m.StringPOD("description")


class ReElementContainer(m.ModelElement, abstract=True):
    elements = m.Containment["CatalogElement"](
        "ownedElements", (NS, "CatalogElement")
    )


class CatalogElementPkg(ReNamedElement, ReElementContainer):
    element_pkgs = m.Containment["CatalogElementPkg"](
        "ownedElementPkgs", (NS, "CatalogElementPkg")
    )


class RecCatalog(CatalogElementPkg):
    compliancy_definition_pkg = m.Containment["CompliancyDefinitionPkg"](
        "ownedCompliancyDefinitionPkg", (NS, "CompliancyDefinitionPkg")
    )


class GroupingElementPkg(CatalogElementPkg):
    pass


class CatalogElementLink(ReAbstractElement):
    source = m.Association["CatalogElement"]((NS, "CatalogElement"), "source")
    target = m.Association["m.ModelElement"](
        (ns.MODELLINGCORE, "ModelElement"), "target"
    )
    origin = m.Association["CatalogElementLink"](
        (NS, "CatalogElementLink"), "origin"
    )
    unsynchronized_features = m.StringPOD("unsynchronizedFeatures")
    is_suffixed = m.BoolPOD("suffixed")


class CatalogElement(ReDescriptionElement, ReElementContainer):
    kind = m.EnumPOD("kind", CatalogElementKind)
    author = m.StringPOD("author")
    environment = m.StringPOD("environment")
    suffix = m.StringPOD("suffix")
    purpose = m.StringPOD("purpose")
    is_read_only = m.BoolPOD("readOnly")
    version = m.StringPOD("version")
    tags = m.StringPOD("tags")
    origin = m.Association["CatalogElement"]((NS, "CatalogElement"), "origin")
    current_compliancy = m.Association["CompliancyDefinition"](
        (NS, "CompliancyDefinition"), "currentCompliancy"
    )
    default_replica_compliancy = m.Association["CompliancyDefinition"](
        (NS, "CompliancyDefinition"), "defaultReplicaCompliancy"
    )
    links = m.Containment["CatalogElementLink"](
        "ownedLinks", (NS, "CatalogElementLink")
    )


class CompliancyDefinitionPkg(ReNamedElement):
    definitions = m.Containment["CompliancyDefinition"](
        "ownedDefinitions", (NS, "CompliancyDefinition")
    )


class CompliancyDefinition(ReDescriptionElement):
    pass



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Tools for handling ReqIF Requirements."""

import typing as t
import warnings

from ._glue import *
from .capellarequirements import *
from .requirements import *

from .capellarequirements import NS as CapellaRequirementsNS  # isort: skip
from .requirements import NS as RequirementsNS  # isort: skip

if not t.TYPE_CHECKING:
    from .capellarequirements import __all__ as _cr_all
    from .requirements import __all__ as _rq_all

    __all__ = [
        "CapellaRequirementsNS",
        "RequirementsNS",
        *_cr_all,
        *_rq_all,
        "AbstractRequirementsAttribute",
        "AbstractRequirementsRelation",
    ]
    del _cr_all, _rq_all

    def __getattr__(name):
        if name == "AbstractRequirementsAttribute":
            warnings.warn(
                "AbstractRequirementsAttribute has been renamed to Attribute",
                FutureWarning,
                stacklevel=2,
            )
            return Attribute

        if name == "AbstractRequirementsRelation":
            warnings.warn(
                "AbstractRequirementsRelation has been renamed to AbstractRelation",
                FutureWarning,
                stacklevel=2,
            )
            return AbstractRelation

        raise AttributeError(f"{__name__} has no attribute {name}")


del t



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

from typing import Protocol

from lxml import etree

class _HasWrite(Protocol):
    def write(self, _: bytes, /) -> None: ...

def serialize(
    tree: etree._Element,
    /,
    *,
    line_length: int,
    siblings: bool,
    file: _HasWrite | None,
) -> bytes: ...



#!/usr/bin/env python
# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import collections
import logging
import re
import subprocess
import sys

import click

from capellambse import helpers

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

CONCOM = re.compile(
    r"^(?P<type>\w+)(?:\((?P<scope>[A-Za-z0-9_-]+)\))?(?P<breaking>\!)?: (?P<subject>.*)$"
)
LOGGER = logging.getLogger(__name__)
MAILUSER = re.compile(
    r"^(?:\d+\+)?(?P<user>[^@]+)@users\.noreply\.github\.com$"
)

HEADINGS: tuple[tuple[str, str | None], ...] = (
    ("BAD", "Commits with malformed subject lines"),
    ("revert", "Reverted earlier changes"),
    ("feat", "New features"),
    ("perf", "Performance improvements"),
    ("fix", "Bug fixes"),
    ("docs", "Documentation changes"),
    ("build", "Build system changes"),
    ("ci", "CI/CD changes"),
    ("test", "Unit test changes"),
    ("refactor", "Code refactorings"),
    ("merge", None),
    ("chore", None),
)


@click.command()
@click.argument("since", required=True)
@click.argument("until", default="HEAD")
def _main(since: str, until: str) -> None:
    """Generate the changelog between the given commit-ish.

    SINCE is exclusive, UNTIL is inclusive.
    """
    try:
        with open("authors.toml", "rb") as f:
            authormap: dict[str, str] = tomllib.load(f)
    except FileNotFoundError:
        LOGGER.warning("authors.toml not found, cannot map authors to users")
        authormap = {}

    format = "format:%s%x00%aE%x00%H"
    all_commits = subprocess.check_output(
        ["git", "log", "-z", f"--format={format}", f"{since}..{until}"],
        encoding="utf-8",
    )
    commits: dict[str, list[str]] = collections.defaultdict(list)
    breaking_changes: list[str] = []

    unknown_authors: set[str] = set()
    for msg, author, hash in helpers.ntuples(3, all_commits.split("\0")):
        if author_match := MAILUSER.search(author):
            author = "@" + author_match.group("user")
        else:
            try:
                author = authormap[author]
            except KeyError:
                if author not in unknown_authors:
                    print(f"Unknown author email: {author}")
                    unknown_authors.add(author)

        if not (msg_match := CONCOM.search(msg)):
            LOGGER.warning("Bad commit subject: %s %s", hash, msg)
            msg = msg[0].upper() + msg[1:]
            commits["BAD"].append(f"{msg} *by {author}* ({hash})")
            continue

        ctype, scope, breaking, subject = msg_match.group(
            "type", "scope", "breaking", "subject"
        )
        scope = f"**{scope}**: " if scope else ""
        subject = subject[0].upper() + subject[1:]
        commits[ctype].append(f"{scope}{subject} *by {author}* ({hash})")
        if breaking:
            breaking_changes.append(f"{scope}{subject}")

    try:
        with open("notable-changes.md", encoding="locale") as f:
            custom_changes = f.read().strip()
    except FileNotFoundError:
        custom_changes = ""
    except Exception as err:
        errtext = f" {type(err).__name__}: {err}"
        custom_changes = f"***Cannot read notable-changes.md:*** {errtext}"

    if breaking_changes:
        print("# Breaking changes", end="\n\n")
        for msg in breaking_changes:
            print(f"- {msg}")
        print()

        print("# Other notable changes", end="\n\n")
    else:
        print("# Notable changes", end="\n\n")
    if custom_changes:
        print(custom_changes.strip(), end="\n\n")
    print("# Full changelog", end="\n\n")

    for ctype, heading in HEADINGS:
        if ctype not in commits:
            continue

        if heading is None:
            del commits[ctype]
            continue

        print(f"## {heading}", end="\n\n")
        for msg in commits.pop(ctype):
            print(f"- {msg}")
        print()

    for ctype, msgs in commits.items():
        print(f"## {ctype.capitalize()}", end="\n\n")
        for msg in msgs:
            print(f"- {msg}")
        print()


if __name__ == "__main__":
    _main()



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import code
import collections.abc as cabc
import contextlib
import importlib
import json
import logging
import operator
import os
import os.path
import pathlib
import shutil
import subprocess
import textwrap
import typing as t

import click
from lxml import etree

import capellambse
from capellambse import cli_helpers
from capellambse.loader import exs

basedir = pathlib.Path(__file__).parent.resolve()
logger = logging.getLogger("capellambse.repl")

try:
    import readline

    with contextlib.suppress(ImportError):
        import rlcompleter  # noqa: F401

    @contextlib.contextmanager
    def _readline_history(histfile: pathlib.Path, /) -> cabc.Iterator[None]:
        try:
            readline.read_history_file(histfile)
        except FileNotFoundError:
            pass
        except OSError:
            backupfile = histfile.with_suffix(".hist.bak")
            logger.exception(
                "Cannot restore prompt history from %r, renaming to %r and starting fresh",
                os.fspath(histfile),
                backupfile.name,
            )
            try:
                histfile.rename(backupfile)
            except OSError as err:
                logger.error(
                    "Could not rename history file, disabling history saving: %s: %s",
                    type(err).__name__,
                    err,
                )
                histfile = pathlib.Path(os.devnull)

        readline.parse_and_bind("tab: complete")

        try:
            yield
        finally:
            try:
                histfile.parent.mkdir(parents=True, exist_ok=True)
                histfile.open("wb").close()
            except OSError:
                pass
            readline.append_history_file(100_000, histfile)

except ImportError:

    @contextlib.contextmanager
    def _readline_history(_histfile, /) -> cabc.Iterator[None]:
        yield


@click.command()
@click.argument("modelinfo", type=cli_helpers.ModelInfoCLI(), required=False)
@click.option(
    "--disable-diagram-cache",
    is_flag=True,
    help="Disable the diagram cache, if one was defined for the model",
)
@click.option(
    "--dump",
    is_flag=True,
    help="Dump model info as JSON to stdout and exit",
)
@click.option(
    "--hold/--no-hold",
    help="Inhibit automatic model updates (update_cache=False)",
)
@click.option(
    "--wipe/--no-wipe",
    help="Wipe the cache (disable_cache=True)",
)
def main(
    *,
    modelinfo: dict[str, t.Any] | None,
    disable_diagram_cache: bool,
    dump: bool,
    hold: bool,
    wipe: bool,
) -> None:
    """Launch a REPL for exploring Capella models.

    You can use this command to quickly launch an interactive Python
    interpreter and load a model.

    If the ``readline`` module is available, the interpreter spawned
    from this script uses a separate readline history. It is located in
    ``$XDG_STATE_HOME/capellambse`` on proper operating systems and in
    the ``capellambse`` cache directory on others.

    It can be used as simply as:

        capellambse repl <MODEL>

    However, while developing capellambse itself, it may be beneficial
    to additionally set some global interpreter flags:

        python -Xdev -Xfrozen_modules=off -m capellambse repl <MODEL>

    This will ensure that development related warnings like
    DeprecationWarning are enabled, and ensure that attached debuggers
    won't miss any breakpoints.

    In addition to the standard Python builtins, the environment in the
    REPL also provides a few convenience imports and helper functions,
    which will be listed right before the first prompt. Further
    information may be obtained using the `help()` builtin.
    """
    os.chdir(pathlib.Path(capellambse.__file__).parents[2])

    if modelinfo is not None:
        if hold:
            modelinfo["update_cache"] = False
        if wipe:
            modelinfo["disable_cache"] = True
        if disable_diagram_cache and "diagram_cache" in modelinfo:
            del modelinfo["diagram_cache"]

    if dump:
        if modelinfo is None:
            raise SystemExit("No model specified, cannot dump")
        print(json.dumps(modelinfo, indent=2))
        raise SystemExit(0)

    interactive_locals = {
        "__doc__": None,
        "__name__": "__console__",
        "etree": etree,
        "fzf": fzf,
        "im": importlib,
        "imm": importlib.import_module("importlib.metadata"),
        "imr": importlib.import_module("importlib.resources"),
        "logtee": logtee,
        "pprint": importlib.import_module("pprint").pprint,
        "showxml": showxml,
        "suppress": suppress,
    }

    for m in ("capellambse", "inspect", "logging", "os", "pathlib"):
        interactive_locals[m] = importlib.import_module(m)

    if modelinfo is not None:
        logger.debug("Loading model: %r", modelinfo["path"])
        model = capellambse.MelodyModel(**modelinfo)
        interactive_locals["model"] = model
        banner = textwrap.dedent(
            f"""\

            {" Model exploration ":=^80}
            `model` is {model.info.title!r}
            from {modelinfo["path"]}
            """
        )
    else:
        banner = textwrap.dedent(
            f"""\

            {" Model exploration ":=^80}
            Load a model with `model = capellambse.loadcli("model-name")`
            or `model = capellambse.MelodyModel("uri", arg=...)`.
            """
        )

    banner += textwrap.dedent(
        """\

        Convenience imports:
        - `capellambse`, `inspect`, `logging`, `os`, `pathlib`
        - `im` = importlib (`imm` = .metadata, `imr` = .resources)
        - `etree` = lxml.etree, `pprint` = pprint.pprint

        Helpful functions and context managers (use `help(name)`):
        - `fzf`: Select a model element interactively from a list
        - `logtee`: CM that redirects log messages to a file
        - `showxml`: Print the XML representation of a model object
        - `suppress`: CM that suppresses exceptions of given type
        """
    )

    history_file = capellambse.dirs.user_state_path / "model_exploration.hist"
    with _readline_history(history_file), suppress(BrokenPipeError):
        code.interact(banner=banner, local=interactive_locals, exitmsg="")


@contextlib.contextmanager
def logtee(
    filename: os.PathLike[str] | str,
    /,
    *,
    append: bool = False,
    catch: bool = False,
    tee: bool = True,
    level: int | str | None = None,
    logger: str = "",
) -> cabc.Generator[None, None, None]:
    """Temporarily write log messages to a file.

    This function adds an additional log handler, meaning that the
    current logging targets (usually stderr) will also still be served.

    Parameters
    ----------
    filename
        The file path to write log messages to.
    append
        Append to the file instead of overwriting it.
    catch
        If ``True``, catches and suppresses exceptions that reach the
        context manager. Only ``Exception`` and subtypes will be
        suppressed. Regardless of this parameter, exceptions (of any
        type) will always be logged (subject to the log level).
    level
        Change the log level during redirection.
    logger
        Attach to this sub-logger instead of the root logger.
    tee
        Keep the original log destination(s) intact as well. If
        ``False``, also inhibits propagation of log messages to parent
        loggers.
    """
    loggerobj = logging.getLogger(logger)
    handler = logging.FileHandler(filename, mode="wa"[bool(append)])

    orig_handlers = loggerobj.handlers
    orig_propagate = loggerobj.propagate
    orig_level = loggerobj.level

    loggerobj.handlers = orig_handlers * bool(tee) + [handler]
    loggerobj.propagate = bool(tee)
    if level is not None:
        loggerobj.setLevel(level)

    try:
        yield
    except BaseException:
        loggerobj.handlers = [handler]
        loggerobj.propagate = False
        loggerobj.exception("Exception reached top of context")
        if not catch:
            raise
    finally:
        loggerobj.handlers = orig_handlers
        loggerobj.propagate = orig_propagate
        loggerobj.level = orig_level
        handler.close()


@contextlib.contextmanager
def suppress(
    *exc_types: type[BaseException], log: bool = True
) -> cabc.Iterator[None]:
    """Prevent the specified types of exceptions from propagating.

    Parameters
    ----------
    exc_types
        The exception types to suppress. Subclasses of these exceptions
        will also be suppressed.
    log
        Print a short warning about the exception to stderr.

    Examples
    --------
    >>> with suppress(ValueError):
    ...     print("Hello")
    ...     raise ValueError("oops")
    ...     print("World")
    ...
    Hello
    """
    try:
        yield
    except exc_types:
        if log:
            logger.info("Exception suppressed", exc_info=True)


def showxml(obj: capellambse.ModelObject | etree._Element) -> None:
    """Show the XML representation of a model object.

    Examples
    --------
    >>> my_obj = model.search("LogicalComponent").by_name("My Component")
    >>> showxml(my_obj)
    <ownedLogicalComponents name="My Component">
      ...
    </ownedLogicalComponents>
    """
    if isinstance(obj, etree._Element):
        elm = obj
    else:
        elm = obj._element
    print(exs.to_string(elm), end="")


def fzf(
    elements: cabc.Iterable[capellambse.ModelObject],
    attr: str = "name",
) -> capellambse.ModelObject | None:
    """Interactively select an element using fzf.

    This function requires the ``fzf`` binary to be installed and in the
    ``$PATH``.

    Examples
    --------
    >>> # Select a LogicalComponent by name
    >>> obj = fzf(model.search("LogicalComponent"))

    >>> # Select a ComponentExchange by the name of its target component
    >>> obj = fzf(model.search("ComponentExchange"), "target.parent.name")
    """

    def repr(obj):
        return getattr(obj, "_short_repr_", obj.__repr__)()

    binary = shutil.which("fzf")
    if not binary:
        raise RuntimeError("fzf is not installed")
    elements = list(elements)

    getter = operator.attrgetter(attr)

    entries = [(str(getter(i)).replace("\0", ""), repr(i)) for i in elements]
    maxlen = max(len(i[0]) for i in entries)
    maxlen = min(maxlen, 40)
    fzf_input = "\0".join(
        f"{i} \x1b[97m{s:{maxlen}}  \x1b[36m{e}"
        for i, (s, e) in enumerate(entries)
    )

    try:
        proc = subprocess.run(
            [binary, "--ansi", "--read0", "--with-nth=2.."],
            check=True,
            input=fzf_input,
            text=True,
            stdout=subprocess.PIPE,
        )
    except (Exception, KeyboardInterrupt):
        return None
    else:
        selected = elements[int(proc.stdout.strip().split(" ", 1)[0])]
        print(repr(selected))
        return selected



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import collections.abc as cabc
import enum
import sys
import typing as t
import warnings

import capellambse.model as m

from . import modellingcore
from . import namespaces as ns

if sys.version_info >= (3, 13):
    from warnings import deprecated
else:
    from typing_extensions import deprecated

NS = ns.CAPELLACORE


@m.stringy_enum
@enum.unique
class VisibilityKind(enum.Enum):
    """The possibilities regarding the visibility of a feature of a class."""

    UNSET = "UNSET"
    """Visibility is not specified."""
    PUBLIC = "PUBLIC"
    """The feature offers public access."""
    PROTECTED = "PROTECTED"
    """The feature offers visibility only to children of the class."""
    PRIVATE = "PRIVATE"
    """The feature is only visible/accessible from the class itself."""
    PACKAGE = "PACKAGE"
    """The feature is accessible from any element within the same package."""


class CapellaElement(
    modellingcore.TraceableElement,
    modellingcore.PublishableElement,
    abstract=True,
):
    summary = m.StringPOD("summary")
    """Summary of the element."""
    description = m.HTMLStringPOD("description")
    """Description of the Capella element."""
    review = m.StringPOD("review")
    """Review description on the Capella element."""

    property_values = m.Containment["AbstractPropertyValue"](
        "ownedPropertyValues",
        (NS, "AbstractPropertyValue"),
        mapkey="name",
        mapvalue="value",
    )
    enumeration_property_types = m.Containment["EnumerationPropertyType"](
        "ownedEnumerationPropertyTypes", (NS, "EnumerationPropertyType")
    )
    applied_property_values = m.Association["AbstractPropertyValue"](
        (NS, "AbstractPropertyValue"), "appliedPropertyValues"
    )
    property_value_groups = m.Containment["PropertyValueGroup"](
        "ownedPropertyValueGroups",
        (NS, "PropertyValueGroup"),
        mapkey="name",
        mapvalue="property_values",
    )
    applied_property_value_groups = m.Association["PropertyValueGroup"](
        (NS, "PropertyValueGroup"), "appliedPropertyValueGroups"
    )
    status = m.Single["EnumerationPropertyLiteral"](
        m.Association((NS, "EnumerationPropertyLiteral"), "status")
    )
    features = m.Association["EnumerationPropertyLiteral"](
        (NS, "EnumerationPropertyLiteral"), "features"
    )


class NamedElement(
    modellingcore.AbstractNamedElement, CapellaElement, abstract=True
):
    pass


class Relationship(
    modellingcore.AbstractRelationship, CapellaElement, abstract=True
):
    pass


class Namespace(NamedElement, abstract=True):
    traces = m.Containment["Trace"]("ownedTraces", (NS, "Trace"))
    naming_rules = m.Containment["NamingRule"](
        "namingRules", (NS, "NamingRule")
    )


class NamedRelationship(Relationship, NamedElement, abstract=True):
    naming_rules = m.Containment["NamingRule"](
        "namingRules", (NS, "NamingRule")
    )


class Structure(Namespace, abstract=True):
    property_value_pkgs = m.Containment["PropertyValuePkg"](
        "ownedPropertyValuePkgs",
        (NS, "PropertyValuePkg"),
        mapkey="name",
    )

    if not t.TYPE_CHECKING:
        property_value_packages = m.DeprecatedAccessor("property_value_pkgs")


class ReuserStructure(Structure, abstract=True):
    reuse_links = m.Association["ReuseLink"]((NS, "ReuseLink"), "reuseLinks")
    owned_reuse_links = m.Containment["ReuseLink"](
        "ownedReuseLinks", (NS, "ReuseLink")
    )


class AbstractModellingStructure(ReuserStructure, abstract=True):
    architectures = m.Containment["ModellingArchitecture"](
        "ownedArchitectures", (NS, "ModellingArchitecture")
    )
    architecture_pkgs = m.Containment["ModellingArchitecturePkg"](
        "ownedArchitecturePkgs", (NS, "ModellingArchitecturePkg")
    )


class Type(modellingcore.AbstractType, Namespace, abstract=True):
    """Represents a set of values.

    A typed element that has this type is constrained to represent
    values within this set.
    """


class ModellingBlock(Type, abstract=True):
    """A modular unit that describes the structure of a system or element.

    A class (or block) that cannot be directly instantiated. Contrast:
    concrete class.
    """


class ModellingArchitecture(Structure, abstract=True):
    """Supports the definition of the model structure at a design level."""


class ModellingArchitecturePkg(Structure, abstract=True):
    """A container for modelling architectures."""


class TypedElement(
    modellingcore.AbstractTypedElement, NamedElement, abstract=True
):
    """An element that has a type."""


class Trace(Relationship, modellingcore.AbstractTrace, abstract=True):
    pass


class AbstractAnnotation(CapellaElement, abstract=True):
    content = m.StringPOD("content")


class NamingRule(AbstractAnnotation):
    target_type = m.StringPOD("targetType")
    """Type to whose instances the naming rule has to be applied."""


class Constraint(NamedElement, modellingcore.AbstractConstraint):
    """A condition or restriction.

    The constraint is expressed in natural language text or in a machine
    readable language for the purpose of declaring some of the semantics
    of an element.
    """

    _xmltag = "ownedConstraints"


class KeyValue(CapellaElement):
    key = m.StringPOD("key")
    value = m.StringPOD("value")


class ReuseLink(Relationship):
    reused = m.Single["ReuseableStructure"](
        m.Association((NS, "ReuseableStructure"), "reused")
    )
    reuser = m.Single["ReuserStructure"](
        m.Association((NS, "ReuserStructure"), "reuser")
    )


class ReuseableStructure(Structure, abstract=True):
    """A structure intended to be reused across various architectures."""

    reuse_links = m.Association["ReuseLink"]((NS, "ReuseLink"), "reuseLinks")


class GeneralizableElement(Type, abstract=True):
    """A type than can be generalized."""

    is_abstract = m.BoolPOD("abstract")
    generalizations = m.Containment["Generalization"](
        "ownedGeneralizations", (NS, "Generalization")
    )
    sub = m.Backref["GeneralizableElement"](
        (NS, "GeneralizableElement"), "super", legacy_by_type=True
    )
    super = m.Single["GeneralizableElement"](
        m.Allocation(
            "ownedGeneralizations",
            (NS, "Generalization"),
            (NS, "GeneralizableElement"),
            attr="super",
            backattr="sub",
        )
    )


class Classifier(GeneralizableElement, abstract=True):
    owned_features = m.Containment["Feature"]("ownedFeatures", (NS, "Feature"))


class GeneralClass(
    Classifier, modellingcore.FinalizableElement, abstract=True
):
    visibility = m.EnumPOD("visibility", VisibilityKind)
    nested_classes = m.Containment["GeneralClass"](
        "nestedGeneralClasses", (NS, "GeneralClass")
    )


class Generalization(Relationship):
    _xmltag = "ownedGeneralizations"

    super = m.Single["GeneralizableElement"](
        m.Association((NS, "GeneralizableElement"), "super")
    )
    sub = m.Single["GeneralizableElement"](
        m.Association((NS, "GeneralizableElement"), "sub")
    )


class Feature(NamedElement, abstract=True):
    is_abstract = m.BoolPOD("isAbstract")
    is_static = m.BoolPOD("isStatic")
    visibility = m.EnumPOD("visibility", VisibilityKind)


class AbstractExchangeItemPkg(Structure, abstract=True):
    exchange_items = m.Containment["information.ExchangeItem"](
        "ownedExchangeItems", (ns.INFORMATION, "ExchangeItem")
    )


class Allocation(Relationship, modellingcore.AbstractTrace, abstract=True):
    pass


class Involvement(Relationship, abstract=True):
    involved = m.Single["InvolvedElement"](
        m.Association((NS, "InvolvedElement"), "involved")
    )

    @property
    @deprecated("Synthetic names are deprecated", category=FutureWarning)
    def name(self) -> str:
        """Return the name."""
        direction = ""
        try:
            if self.involved is not None:
                direction = f" to {self.involved.name} ({self.involved.uuid})"
        except AttributeError:
            pass

        return f"[{self.__class__.__name__}]{direction}"

    if not t.TYPE_CHECKING:
        source = m.DeprecatedAccessor("parent")
        target = m.DeprecatedAccessor("involved")


class InvolverElement(CapellaElement, abstract=True):
    pass


class InvolvedElement(CapellaElement, abstract=True):
    pass


class AbstractPropertyValue(NamedElement, abstract=True):
    _xmltag = "ownedPropertyValues"

    involved_elements = m.Association["CapellaElement"](
        (NS, "CapellaElement"), "involvedElements"
    )

    if not t.TYPE_CHECKING:
        enumerations = m.DeprecatedAccessor("enumeration_property_types")


class StringPropertyValue(AbstractPropertyValue):
    """A string property value."""

    value = m.StringPOD("value")


class IntegerPropertyValue(AbstractPropertyValue):
    """An integer property value."""

    value = m.IntPOD("value")


class BooleanPropertyValue(AbstractPropertyValue):
    """A boolean property value."""

    value = m.BoolPOD("value")


class FloatPropertyValue(AbstractPropertyValue):
    """A floating point property value."""

    value = m.FloatPOD("value")


class EnumerationPropertyValue(AbstractPropertyValue):
    type = m.Single["EnumerationPropertyType"](
        m.Association((NS, "EnumerationPropertyType"), "type")
    )
    value = m.Single["EnumerationPropertyLiteral"](
        m.Association((NS, "EnumerationPropertyLiteral"), "value")
    )


class EnumerationPropertyType(NamedElement):
    _xmltag = "ownedEnumerationPropertyTypes"

    literals = m.Containment["EnumerationPropertyLiteral"](
        "ownedLiterals", (NS, "EnumerationPropertyLiteral")
    )


class EnumerationPropertyLiteral(NamedElement):
    """A Literal for EnumerationPropertyType."""

    _xmltag = "ownedLiterals"


class PropertyValueGroup(Namespace, cabc.MutableMapping[str, str]):
    """A group for PropertyValues."""

    _xmltag = "ownedPropertyValueGroups"

    def __getitem__(self, k: str, /) -> t.Any:
        return self.property_values.by_name(k).value

    def __setitem__(self, k: str, v: t.Any, /) -> None:
        self.property_values.by_name(k).value = v

    def __delitem__(self, k: str, /) -> None:
        self.property_values.delete_all(name=k)

    def __iter__(self) -> cabc.Iterator[str]:
        for prop in self.property_values:
            yield prop.name

    def __len__(self) -> int:
        return sum(1 for _ in self)

    if not t.TYPE_CHECKING:
        values = m.DeprecatedAccessor("property_values")


class PropertyValuePkg(Structure):
    """A Package for PropertyValues."""

    _xmltag = "ownedPropertyValuePkgs"

    packages = m.Alias["m.ElementList[PropertyValuePkg]"](
        "property_value_pkgs", dirhide=False
    )
    groups = m.Alias["m.ElementList[PropertyValueGroup]"](
        "property_value_groups", dirhide=False
    )
    values = m.Alias["m.ElementList[AbstractPropertyValue]"](
        "property_values", dirhide=False
    )


class AbstractDependenciesPkg(Structure, abstract=True):
    pass


if not t.TYPE_CHECKING:

    def __getattr__(attr):
        if attr == "PropertyValue":
            warnings.warn(
                "PropertyValue has been renamed to AbstractPropertyValue",
                DeprecationWarning,
                stacklevel=2,
            )
            return AbstractPropertyValue

        raise AttributeError(f"{__name__} has no attribute {attr}")


from . import information  # noqa: F401



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import enum

import capellambse.model as m
from capellambse.model._obj import ModelElement as ModelElement

from . import namespaces as ns

NS = ns.MODELLINGCORE


@m.stringy_enum
@enum.unique
class ParameterEffectKind(enum.Enum):
    """A behavior's effect on values passed in or out of its parameters."""

    READ = "read"
    """The parameter value is only being read upon behavior execution."""
    UPDATE = "update"
    """The parameter value is being updated upon behavior execution."""
    CREATE = "create"
    """The parameter value is being created upon behavior execution."""
    DELETE = "delete"
    """The parameter value is being deleted upon behavior execution."""


@m.stringy_enum
@enum.unique
class RateKind(enum.Enum):
    """The possible caracterizations for the rate of a streaming parameter."""

    UNSPECIFIED = "Unspecified"
    """The rate kind is not specified."""
    CONTINUOUS = "Continuous"
    """The rate characterizes a continuous flow."""
    DISCRETE = "Discrete"
    """The rate characterizes a discrete flow."""


ModelElement.extensions = m.Containment["ModelElement"](
    "ownedExtensions", (NS, "ModelElement")
)
ModelElement.constraints = m.Containment(
    "ownedConstraints", (ns.CAPELLACORE, "Constraint")
)
ModelElement.migrated_elements = m.Containment["ModelElement"](
    "ownedMigratedElements", (NS, "ModelElement")
)


class AbstractRelationship(ModelElement, abstract=True):
    realized_flow = m.Association["AbstractInformationFlow"](
        (NS, "AbstractInformationFlow"), "realizedFlow"
    )


class AbstractNamedElement(ModelElement, abstract=True):
    """An element that may have a name.

    The name is used for identification of the named element within the
    namespace in which it is defined. A named element also has a
    qualified name that allows it to be unambiguously identified within
    a hierarchy of nested namespaces.
    """

    name = m.StringPOD("name")


class InformationsExchanger(ModelElement, abstract=True):
    """An element that may exchange information with other elements."""


class TraceableElement(ModelElement, abstract=True):
    """An element that may be traced to other elements."""


class FinalizableElement(ModelElement, abstract=True):
    is_final = m.BoolPOD("final")


class PublishableElement(ModelElement, abstract=True):
    is_visible_in_doc = m.BoolPOD("visibleInDoc")
    is_visible_in_lm = m.BoolPOD("visibleInLM")


class AbstractType(AbstractNamedElement, abstract=True):
    """Base abstract class supporting the definition of data types."""


class AbstractTypedElement(AbstractNamedElement, abstract=True):
    """A (named) model element to which a specific type is associated."""

    type = m.Single["AbstractType"](
        m.Association((NS, "AbstractType"), "abstractType")
    )


class AbstractTrace(TraceableElement, abstract=True):
    target = m.Single["TraceableElement"](
        m.Association((NS, "TraceableElement"), "targetElement")
    )
    source = m.Single["TraceableElement"](
        m.Association((NS, "TraceableElement"), "sourceElement")
    )


class AbstractConstraint(ModelElement, abstract=True):
    """A constraint that applies to a given set of model elements."""

    constrained_elements = m.Association["ModelElement"](
        (NS, "ModelElement"), "constrainedElements", legacy_by_type=True
    )
    specification = m.Single["ValueSpecification"](
        m.Containment("ownedSpecification", (NS, "ValueSpecification"))
    )
    """A condition that must evaluate to true to satisfy the constraint."""


class ValueSpecification(AbstractTypedElement, abstract=True):
    """The specification of a set of instances.

    The set includes both objects and data values, and may be empty.
    """


class AbstractParameter(AbstractTypedElement, abstract=True):
    """Specification of an argument to a behavioral feature.

    Parameters are used to pass information into or out of an invocation
    of a behavioral feature.
    """

    is_exception = m.BoolPOD("isException")
    is_stream = m.BoolPOD("isStream")
    is_optional = m.BoolPOD("isOptional")
    kind_of_rate = m.EnumPOD("kindOfRate", RateKind)
    effect = m.EnumPOD("effect", ParameterEffectKind)
    rate = m.Single["ValueSpecification"](
        m.Containment("rate", (NS, "ValueSpecification"))
    )
    probability = m.Single["ValueSpecification"](
        m.Containment("probability", (NS, "ValueSpecification"))
    )
    parameter_set = m.Single["AbstractParameterSet"](
        m.Association((NS, "AbstractParameterSet"), "parameterSet")
    )


class AbstractParameterSet(AbstractNamedElement, abstract=True):
    """An alternative set of inputs or outputs that a behavior may use."""

    conditions = m.Containment["AbstractConstraint"](
        "ownedConditions", (NS, "AbstractConstraint")
    )
    probability = m.Single["ValueSpecification"](
        m.Containment("probability", (NS, "ValueSpecification"))
    )
    parameters = m.Single["AbstractParameter"](
        m.Association((NS, "AbstractParameter"), "parameters")
    )


class AbstractInformationFlow(
    AbstractNamedElement, AbstractRelationship, abstract=True
):
    realizations = m.Association["AbstractRelationship"](
        (NS, "AbstractRelationship"), "realizations"
    )
    convoyed_informations = m.Association["AbstractExchangeItem"](
        (NS, "AbstractExchangeItem"), "convoyedInformations"
    )
    source = m.Single["InformationsExchanger"](
        m.Association((NS, "InformationsExchanger"), "source")
    )
    target = m.Single["InformationsExchanger"](
        m.Association((NS, "InformationsExchanger"), "target")
    )


class AbstractExchangeItem(AbstractType, abstract=True):
    """Set of exchanged element exchanged between ports."""


class IState(AbstractNamedElement, abstract=True):
    """A vertex is an abstraction of a node in a state machine graph.

    In general, it can be the source or destination of any number of
    transitions.
    """

    referenced_states = m.Association["IState"](
        (NS, "IState"), "referencedStates"
    )
    exploited_states = m.Association["IState"](
        (NS, "IState"), "exploitedStates"
    )



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Implementation of objects and relations for Functional Analysis."""

from __future__ import annotations

import typing as t
import warnings

import capellambse.model as m

from . import capellacore, fa, information, modellingcore
from . import namespaces as ns

NS = ns.CS


class BlockArchitecturePkg(
    capellacore.ModellingArchitecturePkg, abstract=True
):
    """Container package for BlockArchitecture elements."""


class BlockArchitecture(fa.AbstractFunctionalArchitecture, abstract=True):
    """Parent class for deriving specific architectures for each design phase.

    Formerly known as BaseArchitectureLayer.
    """

    capability_pkg = m.Single["capellacommon.AbstractCapabilityPkg"](
        m.Containment(
            "ownedAbstractCapabilityPkg",
            (ns.CAPELLACOMMON, "AbstractCapabilityPkg"),
        )
    )
    interface_pkg = m.Single["InterfacePkg"](
        m.Containment("ownedInterfacePkg", (NS, "InterfacePkg"))
    )
    data_pkg = m.Single["information.DataPkg"](
        m.Containment("ownedDataPkg", (ns.INFORMATION, "DataPkg"))
    )

    @property
    def all_classes(self) -> m.ElementList[information.Class]:
        return self._model.search((ns.INFORMATION, "Class"), below=self)

    @property
    def all_collections(self) -> m.ElementList[information.Collection]:
        return self._model.search((ns.INFORMATION, "Collection"), below=self)

    @property
    def all_unions(self) -> m.ElementList[information.Union]:
        return self._model.search((ns.INFORMATION, "Union"), below=self)

    @property
    def all_enumerations(
        self,
    ) -> m.ElementList[information.datatype.Enumeration]:
        return self._model.search(
            (ns.INFORMATION_DATATYPE, "Enumeration"), below=self
        )

    @property
    def all_complex_values(
        self,
    ) -> m.ElementList[information.datavalue.AbstractComplexValue]:
        return self._model.search(
            (ns.INFORMATION_DATAVALUE, "AbstractComplexValue"), below=self
        )

    @property
    def all_interfaces(self) -> m.ElementList[Interface]:
        return self._model.search((NS, "Interface"), below=self)

    @property
    def all_capabilities(
        self,
    ) -> m.ElementList[interaction.AbstractCapability]:
        return self._model.search(
            (ns.INTERACTION, "AbstractCapability"), below=self
        )

    if not t.TYPE_CHECKING:
        capability_package = m.DeprecatedAccessor("capability_pkg")
        interface_package = m.DeprecatedAccessor("interface_pkg")
        data_package = m.DeprecatedAccessor("data_pkg")


class Block(
    fa.AbstractFunctionalBlock,
    # NOTE: In the upstream metamodel, ModellingBlock comes first,
    # but this would result in an MRO conflict.
    capellacore.ModellingBlock,
    abstract=True,
):
    """A modular unit that describes the structure of a system or element."""

    capability_pkg = m.Single["capellacommon.AbstractCapabilityPkg"](
        m.Containment(
            "ownedAbstractCapabilityPkg",
            (ns.CAPELLACOMMON, "AbstractCapabilityPkg"),
        )
    )
    interface_pkg = m.Single["InterfacePkg"](
        m.Containment("ownedInterfacePkg", (NS, "InterfacePkg"))
    )
    data_pkg = m.Single["information.DataPkg"](
        m.Containment("ownedDataPkg", (ns.INFORMATION, "DataPkg"))
    )
    state_machines = m.Containment["capellacommon.StateMachine"](
        "ownedStateMachines", (ns.CAPELLACOMMON, "StateMachine")
    )

    if not t.TYPE_CHECKING:
        interface_package = m.DeprecatedAccessor("interface_pkg")
        data_package = m.DeprecatedAccessor("data_pkg")


class ComponentArchitecture(BlockArchitecture, abstract=True):
    """A specialized kind of BlockArchitecture.

    Serves as a parent class for the various architecture levels, from
    System analysis down to EPBS architecture.
    """


class InterfaceAllocator(capellacore.CapellaElement, abstract=True):
    interface_allocations = m.Containment["InterfaceAllocation"](
        "ownedInterfaceAllocations", (NS, "InterfaceAllocation")
    )
    allocated_interfaces = m.Allocation["Interface"](
        "ownedInterfaceAllocations",
        (NS, "InterfaceAllocation"),
        (NS, "Interface"),
        attr="targetElement",
        backattr="sourceElement",
    )


class Component(
    Block,
    capellacore.Classifier,
    InterfaceAllocator,
    information.communication.CommunicationLinkExchanger,
    abstract=True,
):
    """An entity, with discrete structure within the system.

    Interacts with other Components of the system, thereby contributing
    at its lowest level to the system properties and characteristics.
    """

    is_actor = m.BoolPOD("actor")
    """Indicates if the Component is an Actor."""
    is_human = m.BoolPOD("human")
    """Indicates whether the Component is a Human."""
    interface_uses = m.Containment["InterfaceUse"](
        "ownedInterfaceUses", (NS, "InterfaceUse")
    )
    used_interfaces = m.Allocation["Interface"](
        "ownedInterfaceUses",
        (NS, "InterfaceUse"),
        (NS, "Interface"),
        attr="usedInterface",
    )
    interface_implementations = m.Containment["InterfaceImplementation"](
        "ownedInterfaceImplementations", (NS, "InterfaceImplementation")
    )
    implemented_interfaces = m.Allocation["Interface"](
        "ownedInterfaceImplementations",
        (NS, "InterfaceImplementation"),
        (NS, "Interface"),
        attr="implementedInterfaces",
    )
    component_realizations = m.Containment["ComponentRealization"](
        "ownedComponentRealizations", (NS, "ComponentRealization")
    )
    realized_components = m.Allocation["Component"](
        "ownedComponentRealizations",
        (NS, "ComponentRealization"),
        (NS, "Component"),
        attr="targetElement",
        backattr="sourceElement",
    )
    realizing_components = m.Backref["Component"](
        (NS, "Component"), "realized_components"
    )
    ports = m.Filter["fa.ComponentPort"](
        "owned_features", (ns.FA, "ComponentPort")
    )
    physical_ports = m.Filter["PhysicalPort"](
        "owned_features", (NS, "PhysicalPort")
    )
    parts = m.Backref["Part"]((NS, "Part"), "type")
    physical_paths = m.Containment["PhysicalPath"](
        "ownedPhysicalPath", (NS, "PhysicalPath")
    )
    physical_links = m.Containment["PhysicalLink"](
        "ownedPhysicalLinks", (NS, "PhysicalLink")
    )
    physical_link_categories = m.Containment["PhysicalLinkCategory"](
        "ownedPhysicalLinkCategories", (NS, "PhysicalLinkCategory")
    )
    related_exchanges = m.Backref["fa.ComponentExchange"](
        (ns.FA, "ComponentExchange"),
        "source",
        "source.owner",
        "target",
        "target.owner",
    )

    if not t.TYPE_CHECKING:
        owner = m.DeprecatedAccessor("parent")
        exchanges = m.DeprecatedAccessor("component_exchanges")


class DeployableElement(capellacore.NamedElement, abstract=True):
    """A physical model element intended to be deployed."""


class DeploymentTarget(capellacore.NamedElement, abstract=True):
    """The physical target that will host a deployable element."""


class AbstractPathInvolvedElement(capellacore.InvolvedElement, abstract=True):
    pass


class Part(
    information.AbstractInstance,
    modellingcore.InformationsExchanger,
    DeployableElement,
    DeploymentTarget,
    AbstractPathInvolvedElement,
):
    """A representation of a physical component.

    In SysML, a Part is an owned property of a Block.
    """

    _xmltag = "ownedParts"

    deployment_links = m.Containment["AbstractDeploymentLink"](
        "ownedDeploymentLinks", (NS, "AbstractDeploymentLink")
    )
    deployed_parts = m.Allocation["DeployableElement"](
        "ownedDeploymentLinks",
        (NS, "AbstractDeploymentLink"),
        (NS, "DeployableElement"),
        attr="deployedElement",
        backattr="location",
    )
    owned_type = m.Single["modellingcore.AbstractType"](
        m.Containment("ownedAbstractType", (ns.MODELLINGCORE, "AbstractType"))
    )


class ArchitectureAllocation(capellacore.Allocation, abstract=True):
    pass


class ComponentRealization(capellacore.Allocation):
    """A realization that links to a component."""

    _xmltag = "ownedComponentRealizations"


class InterfacePkg(
    information.communication.MessageReferencePkg,
    capellacore.AbstractDependenciesPkg,
    capellacore.AbstractExchangeItemPkg,
):
    """A container for Interface elements."""

    interfaces = m.Containment["Interface"](
        "ownedInterfaces", (NS, "Interface")
    )
    packages = m.Containment["InterfacePkg"](
        "ownedInterfacePkgs", (NS, "InterfacePkg")
    )


class Interface(capellacore.GeneralClass, InterfaceAllocator):
    """An interface.

    An interface is a kind of classifier that represents a declaration of a set
    of coherent public features and obligations. An interface specifies a
    contract; any instance of a classifier that realizes the interface must
    fulfill that contract.

    Interfaces are defined by functional and physical characteristics that
    exist at a common boundary with co-functioning items and allow systems,
    equipment, software, and system data to be compatible.

    That design feature of one piece of equipment that affects a design feature
    of another piece of equipment. An interface can extend beyond the physical
    boundary between two items. (For example, the weight and center of gravity
    of one item can affect the interfacing item; however, the center of gravity
    is rarely located at the physical boundary. An electrical interface
    generally extends to the first isolating element rather than terminating at
    a series of connector pins.)

    Usage guideline
    ---------------
    In Capella, Interfaces are created to declare the nature of interactions
    between the System and external actors.
    """

    mechanism = m.StringPOD("mechanism")
    is_structural = m.BoolPOD("structural")
    exchange_item_allocations = m.Containment["ExchangeItemAllocation"](
        "ownedExchangeItemAllocations", (NS, "ExchangeItemAllocation")
    )
    allocated_exchange_items = m.Allocation["information.ExchangeItem"](
        "ownedExchangeItemAllocations",
        (NS, "ExchangeItemAllocation"),
        (ns.INFORMATION, "ExchangeItem"),
        attr="allocatedItem",
    )


class InterfaceImplementation(capellacore.Relationship):
    implemented_interface = m.Single["Interface"](
        m.Association((NS, "Interface"), "implementedInterface")
    )


class InterfaceUse(capellacore.Relationship):
    used_interface = m.Single["Interface"](
        m.Association((NS, "Interface"), "usedInterface")
    )


class ProvidedInterfaceLink(capellacore.Relationship, abstract=True):
    interface = m.Single["Interface"](
        m.Association((NS, "Interface"), "interface")
    )


class RequiredInterfaceLink(capellacore.Relationship, abstract=True):
    interface = m.Single["Interface"](
        m.Association((NS, "Interface"), "interface")
    )


class InterfaceAllocation(capellacore.Allocation, abstract=True):
    pass


class ExchangeItemAllocation(
    capellacore.Relationship,
    information.AbstractEventOperation,
    modellingcore.FinalizableElement,
):
    """An allocation of an ExchangeItem to an Interface."""

    send_protocol = m.EnumPOD(
        "sendProtocol", information.communication.CommunicationLinkProtocol
    )
    receive_protocol = m.EnumPOD(
        "receiveProtocol", information.communication.CommunicationLinkProtocol
    )
    allocated_item = m.Single["information.ExchangeItem"](
        m.Association((ns.INFORMATION, "ExchangeItem"), "allocatedItem")
    )

    if not t.TYPE_CHECKING:
        item = m.DeprecatedAccessor("allocated_item")


class AbstractDeploymentLink(capellacore.Relationship, abstract=True):
    deployed_element = m.Single["DeployableElement"](
        m.Association((NS, "DeployableElement"), "deployedElement")
    )
    location = m.Single["DeploymentTarget"](
        m.Association((NS, "DeploymentTarget"), "location")
    )


class AbstractPhysicalArtifact(capellacore.CapellaElement, abstract=True):
    pass


class AbstractPhysicalLinkEnd(capellacore.CapellaElement, abstract=True):
    pass


class AbstractPhysicalPathLink(fa.ComponentExchangeAllocator, abstract=True):
    pass


class PhysicalLink(
    AbstractPhysicalPathLink,
    AbstractPhysicalArtifact,
    AbstractPathInvolvedElement,
):
    ends = m.Association["AbstractPhysicalLinkEnd"](
        (NS, "AbstractPhysicalLinkEnd"), "linkEnds", fixed_length=2
    )
    functional_exchange_allocations = m.Containment[
        "fa.ComponentExchangeFunctionalExchangeAllocation"
    ](
        "ownedComponentExchangeFunctionalExchangeAllocations",
        (ns.FA, "ComponentExchangeFunctionalExchangeAllocation"),
    )
    allocated_functional_exchanges = m.Allocation["fa.FunctionalExchange"](
        "ownedComponentExchangeFunctionalExchangeAllocations",
        (ns.FA, "ComponentExchangeFunctionalExchangeAllocation"),
        (ns.FA, "FunctionalExchange"),
        attr="targetElement",
        backattr="sourceElement",
    )
    physical_link_ends = m.Containment["PhysicalLinkEnd"](
        "ownedPhysicalLinkEnds", (NS, "PhysicalLinkEnd")
    )
    physical_link_realizations = m.Containment["PhysicalLinkRealization"](
        "ownedPhysicalLinkRealizations", (NS, "PhysicalLinkRealization")
    )
    realized_physical_links = m.Allocation["PhysicalLink"](
        "ownedPhysicalLinkRealizations",
        (NS, "PhysicalLinkRealization"),
        (NS, "PhysicalLink"),
        attr="targetElement",
        backattr="sourceElement",
    )
    physical_paths = m.Backref["PhysicalPath"](
        (NS, "PhysicalPath"), "involved_items"
    )

    @property
    def source(self) -> AbstractPhysicalLinkEnd | None:
        try:
            return self.ends[0]
        except IndexError:
            return None

    @source.setter
    def source(self, end: AbstractPhysicalLinkEnd | None) -> None:
        if end is None:
            raise TypeError(f"Cannot delete 'source' of {type(self).__name__}")

        ends = self.ends
        if len(ends) == 0:
            ends.append(end)
        else:
            ends[0] = end

    @property
    def target(self) -> AbstractPhysicalLinkEnd | None:
        try:
            return self.ends[1]
        except IndexError:
            return None

    @target.setter
    def target(self, end: AbstractPhysicalLinkEnd | None) -> None:
        if end is None:
            raise TypeError(f"Cannot delete 'target' of {type(self).__name__}")

        ends = self.ends
        if len(ends) == 0:
            raise TypeError(
                f"Cannot set 'target' on a {type(self).__name__}"
                " that has no 'source'"
            )
        if len(ends) == 1:
            ends.append(end)
        else:
            ends[1] = end

    def links(self) -> m.ElementList[m.ModelElement]:
        warnings.warn(
            "PhysicalLink.links is deprecated and will be removed soon",
            category=FutureWarning,
            stacklevel=2,
        )
        return m.ElementList(self._model, [])

    if not t.TYPE_CHECKING:
        exchanges = m.DeprecatedAccessor("allocated_component_exchanges")
        owner = m.DeprecatedAccessor("parent")


class PhysicalLinkCategory(capellacore.NamedElement):
    links = m.Association["PhysicalLink"]((NS, "PhysicalLink"), "links")


class PhysicalLinkEnd(AbstractPhysicalLinkEnd):
    port = m.Single["PhysicalPort"](
        m.Association((NS, "PhysicalPort"), "port")
    )
    part = m.Single["Part"](m.Association((NS, "Part"), "part"))


class PhysicalLinkRealization(capellacore.Allocation):
    pass


class PhysicalPath(
    fa.ComponentExchangeAllocator,
    AbstractPathInvolvedElement,
    capellacore.InvolverElement,
    capellacore.NamedElement,
):
    """A physical path."""

    _xmltag = "ownedPhysicalPath"

    _involved_links = m.Association["AbstractPhysicalPathLink"](
        (NS, "AbstractPhysicalPathLink"), "involvedLinks"
    )
    physical_path_involvements = m.Containment["PhysicalPathInvolvement"](
        "ownedPhysicalPathInvolvements", (NS, "PhysicalPathInvolvement")
    )
    involved_items = m.Allocation["AbstractPathInvolvedElement"](
        "ownedPhysicalPathInvolvements",
        (NS, "PhysicalPathInvolvement"),
        (NS, "AbstractPathInvolvedElement"),
        attr="involved",
        legacy_by_type=True,
    )
    physical_path_realizations = m.Containment["PhysicalPathRealization"](
        "ownedPhysicalPathRealizations", (NS, "PhysicalPathRealization")
    )
    realized_paths = m.Allocation["PhysicalPath"](
        "ownedPhysicalPathRealizations",
        (NS, "PhysicalPathRealization"),
        (NS, "PhysicalPath"),
        attr="targetElement",
        backattr="sourceElement",
    )
    involved_links = m.Filter["PhysicalLink"](
        "involved_items", (NS, "PhysicalLink")
    )

    if not t.TYPE_CHECKING:
        exchanges = m.DeprecatedAccessor("allocated_component_exchanges")


class PhysicalPathInvolvement(capellacore.Involvement):
    next_involvements = m.Association["PhysicalPathInvolvement"](
        (NS, "PhysicalPathInvolvement"), "nextInvolvements"
    )


class PhysicalPathReference(PhysicalPathInvolvement):
    pass


class PhysicalPathRealization(capellacore.Allocation):
    pass


class PhysicalPort(
    information.Port,
    AbstractPhysicalArtifact,
    modellingcore.InformationsExchanger,
    AbstractPhysicalLinkEnd,
    information.Property,
):
    """A port on a physical component."""

    _xmltag = "ownedFeatures"

    component_port_allocations = m.Containment["fa.ComponentPortAllocation"](
        "ownedComponentPortAllocations",
        (ns.FA, "ComponentPortAllocation"),
    )
    allocated_component_ports = m.Allocation["fa.ComponentPort"](
        "ownedComponentPortAllocations",
        (ns.FA, "ComponentPortAllocation"),
        (ns.FA, "ComponentPort"),
        attr="targetElement",
        backattr="sourceElement",
    )
    physical_port_realizations = m.Containment["PhysicalPortRealization"](
        "ownedPhysicalPortRealizations", (NS, "PhysicalPortRealization")
    )
    realized_ports = m.Allocation["PhysicalPort"](
        "ownedPhysicalPortRealizations",
        (NS, "PhysicalPortRealization"),
        (NS, "PhysicalPort"),
        attr="targetElement",
        backattr="sourceElement",
    )
    links = m.Backref["PhysicalLink"]((NS, "PhysicalLink"), "ends")

    if not t.TYPE_CHECKING:
        owner = m.DeprecatedAccessor("parent")


class PhysicalPortRealization(capellacore.Allocation):
    pass


class ComponentPkg(capellacore.Structure, abstract=True):
    """A package containing parts."""

    parts = m.Containment["Part"]("ownedParts", (NS, "Part"))
    exchanges = m.Containment["fa.ComponentExchange"](
        "ownedComponentExchanges", (ns.FA, "ComponentExchange")
    )
    exchange_categories = m.Containment["fa.ComponentExchangeCategory"](
        "ownedComponentExchangeCategories",
        (ns.FA, "ComponentExchangeCategory"),
    )
    functional_links = m.Containment["fa.ExchangeLink"](
        "ownedFunctionalLinks", (ns.FA, "ExchangeLink")
    )
    functional_allocations = m.Containment["fa.ComponentFunctionalAllocation"](
        "ownedFunctionalAllocations",
        (ns.FA, "ComponentFunctionalAllocation"),
    )
    allocated_functions = m.Allocation["fa.AbstractFunction"](
        "ownedFunctionalAllocations",
        (ns.FA, "ComponentFunctionalAllocation"),
        (ns.FA, "AbstractFunction"),
        attr="targetElement",
        backattr="sourceElement",
    )
    component_exchange_realizations = m.Containment[
        "fa.ComponentExchangeRealization"
    ](
        "ownedComponentExchangeRealizations",
        (ns.FA, "ComponentExchangeRealization"),
    )
    realized_component_exchanges = m.Allocation["fa.ComponentExchange"](
        "ownedComponentExchangeRealizations",
        (ns.FA, "ComponentExchangeRealization"),
        (ns.FA, "ComponentExchange"),
        attr="targetElement",
        backattr="sourceElement",
    )
    physical_links = m.Containment["PhysicalLink"](
        "ownedPhysicalLinks", (NS, "PhysicalLink")
    )
    physical_link_categories = m.Containment["PhysicalLinkCategory"](
        "ownedPhysicalLinkCategories", (NS, "PhysicalLinkCategory")
    )
    state_machines = m.Containment["capellacommon.StateMachine"](
        "ownedStateMachines", (ns.CAPELLACOMMON, "StateMachine")
    )


from . import capellacommon, interaction  # noqa: F401



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""The capellambse package."""

import platformdirs

dirs = platformdirs.PlatformDirs("capellambse")
del platformdirs

from importlib import metadata

try:
    __version__ = metadata.version("capellambse")
except metadata.PackageNotFoundError:
    __version__ = "0.0.0+unknown"
del metadata

from ._namespaces import *
from .cli_helpers import *
from .filehandler import *
from .model import MelodyModel as MelodyModel
from .model import ModelObject as ModelObject
from .model import NewObject as NewObject

_has_loaded_extensions = False


def load_model_extensions() -> None:
    """Load all model extensions.

    This function loads all entry points in the group
    ``capellambse.model_extensions`` and executes them.

    It is automatically called when loading a model. Calling it more
    than once has no effect, so it is safe (although not necessary) to
    explicitly call this function before loading a model.
    """
    # Reduce namespace pollution
    import importlib.metadata as imm  # noqa: PLC0415
    import logging  # noqa: PLC0415

    global _has_loaded_extensions
    if _has_loaded_extensions:
        return
    _has_loaded_extensions = True

    for entrypoint in imm.entry_points(group="capellambse.model_extensions"):
        try:
            initfunc = entrypoint.load()
            initfunc()
        except Exception:
            logging.getLogger(__name__).exception(
                "Cannot load model extension %r from %r",
                entrypoint.name,
                entrypoint.value,
            )



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

import typing as t

import click

import capellambse
from capellambse import decl


@click.command()
@click.option("-m", "--model", type=capellambse.ModelCLI(), required=True)
@click.option("-s", "--strict/--relaxed", is_flag=True, default=False)
@click.argument("file", type=click.File("r"))
def main(
    model: capellambse.MelodyModel,
    file: t.IO[str],
    strict: bool,
) -> None:
    """Apply a declarative modelling YAML file to a model."""
    decl.apply(model, file, strict=strict)
    model.save()



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

__all__ = [
    "Category",
    "ElementValidation",
    "LayerValidation",
    "ModelValidation",
    "RealType",
    "Result",
    "Results",
    "Rule",
    "Rules",
    "Validation",
    "VirtualType",
    "rule",
    "virtual_type",
]

import collections.abc as cabc
import dataclasses
import enum
import logging
import sys
import typing as t

from lxml import etree

import capellambse
import capellambse.metamodel as mm
import capellambse.model as m

if sys.version_info >= (3, 13):
    from warnings import deprecated
else:
    from typing_extensions import deprecated

LOGGER = logging.getLogger(__name__)
_T_co = t.TypeVar(
    "_T_co", covariant=True, bound=t.Union[m.ModelElement, "VirtualType"]
)


@dataclasses.dataclass(frozen=True)
class VirtualType(t.Generic[_T_co]):
    name: str
    real_type: type[_T_co]
    filter: cabc.Callable[[_T_co], bool]

    def search(self, model_: capellambse.MelodyModel) -> m.ElementList:
        assert isinstance(self.real_type, str | type(m.ModelElement))
        return model_.search(self.real_type).filter(self.filter)

    def matches(self, obj: m.ModelElement) -> bool:
        return isinstance(obj, self.real_type) and self.filter(obj)


@dataclasses.dataclass(frozen=True)
class RealType(t.Generic[_T_co]):
    class_: m.ClassName

    @property
    def name(self) -> str:
        return self.class_[1]

    def search(self, model_: capellambse.MelodyModel) -> m.ElementList:
        return model_.search(self.class_)

    def matches(self, obj: m.ModelElement) -> bool:
        return (
            obj.__capella_namespace__ is self.class_[0]
            and type(obj).__name__ == self.class_[1]
        )


class _VirtualTypesRegistry(cabc.Mapping[str, VirtualType | RealType]):
    def __init__(self) -> None:
        self.__registry: dict[str, VirtualType] = {}

    def __iter__(self) -> cabc.Iterator[str]:
        yield from self.__registry

    def __len__(self) -> int:
        return len(self.__registry)

    def __contains__(self, key: object) -> bool:
        return key in self.__registry

    def __getitem__(self, key: str) -> VirtualType | RealType:
        try:
            return self.__registry[key]
        except KeyError:
            pass
        return RealType(m.resolve_class_name(("", key)))

    def register(self, vtype: VirtualType) -> None:
        try:
            known = self.__registry[vtype.name]
        except KeyError:
            pass
        else:
            raise RuntimeError(f"Virtual type already known: {known}")
        self.__registry[vtype.name] = vtype


_types_registry = _VirtualTypesRegistry()


def virtual_type(
    real_type: str | type[_T_co],
) -> cabc.Callable[[cabc.Callable[[_T_co], bool]], VirtualType[_T_co]]:
    if isinstance(real_type, str):
        (cls,) = t.cast(tuple[type[_T_co], ...], m.find_wrapper(real_type))
    else:
        cls = real_type

    def decorate(func: cabc.Callable[[_T_co], bool]) -> VirtualType[_T_co]:
        vtype = VirtualType(func.__name__, cls, func)
        _types_registry.register(vtype)
        return vtype

    return decorate


@virtual_type(mm.oa.OperationalActivity)
def OperationalActivity(obj):
    assert isinstance(obj, mm.oa.OperationalActivity)
    assert hasattr(obj._model, "oa"), "Model doesn't have an OA layer?"
    pkg = obj._model.oa.activity_pkg
    assert pkg is not None
    return obj not in pkg.activities


@virtual_type(mm.sa.SystemFunction)
def SystemFunction(obj):
    assert hasattr(obj._model, "sa"), "Model doesn't have an SA layer?"
    return obj != obj._model.sa.root_function


@virtual_type(mm.la.LogicalFunction)
def LogicalFunction(obj):
    assert hasattr(obj._model, "sa"), "Model doesn't have an LA layer?"
    return obj != obj._model.la.root_function


@virtual_type(mm.pa.PhysicalFunction)
def PhysicalFunction(obj):
    assert hasattr(obj._model, "sa"), "Model doesn't have a PA layer?"
    return obj != obj._model.pa.root_function


@m.stringy_enum
class Category(enum.Enum):
    """A category for a rule."""

    REQUIRED = enum.auto()
    RECOMMENDED = enum.auto()
    SUGGESTED = enum.auto()


@dataclasses.dataclass
class Result:
    """The result of checking a validation rule against a model object."""

    rule: Rule
    object: m.ModelElement
    passed: bool

    def __repr__(self) -> str:
        """Return the representation of a result."""
        return (
            f"{type(self).__name__}(rule={self.rule!r},"
            f" object={self.object._short_repr_()},"
            f" passed={self.passed!r})"
        )


@dataclasses.dataclass(frozen=True)
class Rule(t.Generic[_T_co]):
    """A validation rule."""

    id: str
    name: str
    types: frozenset[str]
    rationale: str
    category: Category
    action: str
    validate: cabc.Callable[[_T_co], bool]

    def find_objects(
        self, model_: capellambse.MelodyModel
    ) -> cabc.Iterator[_T_co]:
        seen: set[str] = set()
        for i in self.types:
            for obj in _types_registry[i].search(model_):
                if obj.uuid in seen:
                    continue
                seen.add(obj.uuid)
                yield obj

    def applies_to(self, obj: m.ModelElement) -> bool:
        """Check whether this Rule applies to a specific element."""
        return any(_types_registry[i].matches(obj) for i in self.types)


class Rules(dict[str, Rule]):
    """Stores validation rules indexed by their ID."""

    def by_category(self, category: Category | str) -> list[Rule]:
        """Filter the validation rules by category."""
        if isinstance(category, str):
            category = Category[category]
        return [i for i in self.values() if i.category == category]

    @deprecated("by_type has been renamed to by_class")
    def by_type(self, type: type[m.ModelElement] | str) -> list[Rule]:
        return self.by_class(type)

    def by_class(self, type: type[m.ModelElement] | str) -> list[Rule]:
        """Filter the validation rules by type."""
        if not isinstance(type, str):
            type = type.__name__
        return [i for i in self.values() if type in i.types]


_VALIDATION_RULES = Rules()


class Results:
    """A set of validation results."""

    def __init__(
        self,
        results: cabc.Iterable[tuple[tuple[Rule, str], Result]] = (),
    ) -> None:
        self.__container = dict(results)

    def __len__(self) -> int:
        return len(self.__container)

    def __iter__(self) -> cabc.Iterator[Result]:
        return iter(self.__container.values())

    def get_result(
        self, rule_: Rule | str, target: str | m.ModelElement, /
    ) -> Result | None:
        if isinstance(rule_, str):
            rule_ = _VALIDATION_RULES[rule_]
        if not isinstance(target, str):
            target = target.uuid
        assert isinstance(target, str)

        try:
            return self.__container[rule_, target]
        except KeyError:
            return None

    def iter_rules(self) -> cabc.Iterator[Rule]:
        seen: set[str] = set()
        for result in self.__container.values():
            if result.rule.id in seen:
                continue
            seen.add(result.rule.id)
            yield result.rule

    def iter_objects(self) -> cabc.Iterator[m.ModelElement]:
        seen: set[str] = set()
        for result in self.__container.values():
            if result.object.uuid in seen:
                continue
            seen.add(result.object.uuid)
            yield result.object

    def iter_results(self) -> cabc.Iterator[Result]:
        return iter(self)

    def iter_compliant_objects(self) -> cabc.Iterator[m.ModelElement]:
        for obj in self.iter_objects():
            if all(i.passed for i in self.by_object(obj).iter_results()):
                yield obj

    def by_rule(self, key: Rule | str, /) -> Results:
        if not isinstance(key, Rule):
            key = _VALIDATION_RULES[key]
        return Results(
            ((rule_, objid), result)
            for (rule_, objid), result in self.__container.items()
            if rule_ == key
        )

    def by_object(self, target: str | m.ModelElement, /) -> Results:
        """Filter the validation results by the target object."""
        if not isinstance(target, str):
            target = target.uuid
        return Results(
            ((rule_, objid), result)
            for (rule_, objid), result in self.__container.items()
            if target == objid
        )

    def by_category(self, category: Category | str, /) -> Results:
        """Filter the validation results by category."""
        return Results(
            ((rule_, objid), result)
            for (rule_, objid), result in self.__container.items()
            if rule_.category == category
        )

    def by_passed(self, passed: bool, /) -> Results:
        """Filter the validation results by whether the rule passed or not."""
        return Results(
            ((rule_, objid), result)
            for (rule_, objid), result in self.__container.items()
            if result.passed == passed
        )

    @deprecated("by_type has been renamed to by_class")
    def by_type(self, /, *types: str) -> Results:
        return self.by_class(*types)

    def by_class(self, /, *types: str) -> Results:
        """Filter the validation results by target object type."""
        if not types:
            raise TypeError("Results.by_class requires at least one argument")
        typeobjs = [_types_registry[i] for i in types]
        return Results(
            ((rule_, objid), result)
            for (rule_, objid), result in self.__container.items()
            if any(t.matches(result.object) for t in typeobjs)
        )


def rule(
    id: str,
    category: Category,
    *,
    name: str,
    rationale: str,
    action: str,
    types: (
        str
        | VirtualType[m.ModelElement]
        | type[m.ModelElement]
        | cabc.Iterable[
            str | VirtualType[m.ModelElement] | type[m.ModelElement]
        ]
    ),
) -> cabc.Callable[[cabc.Callable[[m.ModelElement], bool]], Rule]:
    """Create a validation rule.

    This decorator registers the validator function as a modelling rule.
    The function is called with each model object to be validated. It
    must return True if the rule passed, and False if it did not pass.

    Parameters
    ----------
    category
        The category of severity for this rule.
    types
        Types of objects that this rule applies to.

        Object types can be either real types (subclasses of
        :class:`~capellambse.model.common.ModelElement`) or virtual
        types created with the :func:`virtual_type` decorator. These can
        be freely mixed in the rule decorator.
    id
        The unique ID of this rule.

        If another rule with this ID already exists, an error is raised.
    name
        Human-readable short name for the rule.
    rationale
        Text describing why the rule is useful.
    action
        Human-reabale short description of what needs to be changed for
        the rule to pass.
    """
    if id in _VALIDATION_RULES:
        raise ValueError(f"Duplicate rule ID: {id}")

    if not types:
        raise TypeError("No 'types' specified")
    if isinstance(types, type):
        type_names = [types.__name__]
    elif isinstance(types, RealType | VirtualType):
        type_names = [types.name]
    elif isinstance(types, str):
        type_names = [types]
    else:
        type_names = []
        for i in types:
            if isinstance(i, str):
                type_names.append(i)
            elif isinstance(i, RealType | VirtualType):
                type_names.append(i.name)
            else:
                type_names.append(i.__name__)

    def rule_decorator(
        validator: cabc.Callable[[m.ModelElement], bool], /
    ) -> Rule:
        rule_ = Rule(
            id,
            name,
            frozenset(type_names),
            rationale,
            category,
            action,
            validator,
        )
        _VALIDATION_RULES.setdefault(rule_.id, rule_)
        assert _VALIDATION_RULES[rule_.id] is rule_
        return rule_

    return rule_decorator


class ModelValidation:
    """Provides access to the model's validation rules and results."""

    _model: capellambse.MelodyModel

    def __init__(self, model: capellambse.MelodyModel) -> None:
        self._model = model

    @property
    def rules(self) -> Rules:
        """Return all registered validation rules."""
        return _VALIDATION_RULES

    def validate(self) -> Results:
        """Execute all registered validation rules and store results."""
        all_results = []
        for rule_ in _VALIDATION_RULES.values():
            for obj in rule_.find_objects(self._model):
                all_results.append(
                    (
                        (rule_, obj.uuid),
                        Result(rule_, obj, rule_.validate(obj)),
                    )
                )
        return Results(all_results)

    def search(self, /, *typenames: str) -> m.ElementList[t.Any]:
        found: dict[str, t.Any] = {}
        for i in typenames:
            objs = _types_registry[i].search(self._model)
            found.update((o.uuid, o._element) for o in objs)
        return m.ElementList(
            self._model, list(found.values()), legacy_by_type=True
        )


class ObjectValidation:
    """Basic class for access to validation rules and results."""

    _model: capellambse.MelodyModel
    _element: etree._Element

    parent = m.AlternateAccessor(m.ModelElement)

    def __init__(self, **kw: t.Any) -> None:
        del kw
        raise TypeError("Cannot create Validation object this way")

    @classmethod
    def from_model(
        cls, model_: capellambse.MelodyModel, element: etree._Element
    ) -> Validation:
        """Create a Validation object for a MelodyModel or an element."""
        self = cls.__new__(cls)
        self._model = model_
        self._element = element
        return self


class LayerValidation(ObjectValidation):
    """Provides access to the layer's validation rules and results."""

    @property
    def rules(self) -> Rules:
        """Return all registered validation rules."""
        return _VALIDATION_RULES

    def validate(self) -> Results:
        """Execute all registered validation rules and store results."""
        all_results = []
        for rule_ in _VALIDATION_RULES.values():
            for obj in rule_.find_objects(self._model):
                if obj.layer != self.parent:
                    continue
                all_results.append(
                    (
                        (rule_, obj.uuid),
                        Result(rule_, obj, rule_.validate(obj)),
                    )
                )
        return Results(all_results)

    def search(self, /, *typenames: str) -> m.ElementList[t.Any]:
        found: dict[str, t.Any] = {}
        for i in typenames:
            objs = _types_registry[i].search(self._model).by_layer(self.parent)
            found.update((o.uuid, o._element) for o in objs)
        return m.ElementList(
            self._model, list(found.values()), legacy_by_type=True
        )


class ElementValidation(ObjectValidation):
    """Provides access to the element's validation rules and results."""

    @property
    def rules(self) -> list[Rule]:
        """Return all registered validation rules that apply to this object."""
        try:
            obj = self.parent
            return [i for i in _VALIDATION_RULES.values() if i.applies_to(obj)]
        except AttributeError as err:
            raise RuntimeError("Cannot compute applicable rules") from err

    def validate(self) -> Results:
        """Validate this element against the rules that apply to it."""
        obj = self.parent
        all_results = []
        for rule_ in _VALIDATION_RULES.values():
            if rule_.applies_to(obj):
                all_results.append(
                    (
                        (rule_, obj.uuid),
                        Result(rule_, obj, rule_.validate(obj)),
                    )
                )
        return Results(all_results)


Validation = ModelValidation | ObjectValidation



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

import json

import pytest

import capellambse
from capellambse import aird, diagram, loader

from .conftest import Models  # type: ignore


class TestAIRDBasicFunctionality:
    (test_model,) = Models.aird_parser.glob("*.aird")
    test_diagram = "[LAB] Logical System"

    @pytest.fixture
    def model(self):
        return loader.MelodyLoader(self.test_model)

    @pytest.fixture
    def diagram_under_test(self, model):
        descriptor = next(aird.enumerate_descriptors(model), None)
        assert descriptor is not None
        return aird.parse_diagram(model, descriptor)

    def test_parsing_all_diagrams_does_not_raise_exceptions(
        self, model, caplog
    ):
        del caplog
        num = sum(1 for _ in aird.parse_diagrams(model))
        assert num == 1

    def test_enumerate_descriptors_doesnt_crash_if_there_are_no_diagrams(self):
        model = loader.MelodyLoader(Models.empty)
        for _ in aird.enumerate_descriptors(model):
            pass


def test_airdparser_msm_produces_valid_json_without_error(
    model: capellambse.MelodyModel,
):
    dg = model.diagrams.by_name("[MSM] States of Functional Human Being")
    parsed = aird.parse_diagram(model._loader, dg._element)

    generated_json = diagram.DiagramJSONEncoder(indent=4).encode(parsed)
    json.loads(generated_json)


@pytest.mark.parametrize(
    ("diag_uid", "num_nodes"),
    [
        ("_7Ft7QKrxEeqOgqWuHJrXFA", 34),
        ("_KLGoEKyJEeqCdMaqCWkrKg", 14),
        ("_be95kNYrEeqiU8uzTY0Puw", 11),
        ("_-NU3gV2mEey8erljvkS-pQ", 13),
        ("_KZRQIWA9Eey8erljvkS-pQ", 15),
        ("_jofb8WDdEey8erljvkS-pQ", 5),
        ("_bpuRwWDhEey8erljvkS-pQ", 4),
        ("_OAjYgXPIEeyW3OIB4qRWZA", 18),
        ("_RSYVoXPIEeyW3OIB4qRWZA", 8),
        ("_bNNAAXPIEeyW3OIB4qRWZA", 6),
        ("_sCBdcMn-EeyS2Zr7ZWFrXA", 7),
        ("_1aihhGpbEe2Scdb2k-Fvhg", 18),
        ("_kqj3UF9REe2rko4oG1H6IQ", 5),
        ("_fr69QBKnEeuBCogvtwwNBw", 43),
        ("_erzycVIqEeyiRNlyKPJwqw", 44),
        ("_8BQYEYQzEeyxv9w6U6_UQg", 6),
        ("_L8b8cCoPEe6CTqm4KXEZ8w", 20),
        ("_amKHoKg0EeujRPrkuugYGw", 5),
        ("_APOQ0QPhEeynfbzU12yy7w", 11),
        ("_K-BhAIKBEeyd4IW8P6K-AQ", 14),
        ("_yov6wM-ZEeytxdoVf3xHjA", 7),
        ("_hJD6gaWXEeqh-bahrg4jkA", 18),
        ("_u8_YMLB-EeqP7JXhmLvOsg", 26),
        ("_2TLuoAL8EeuogvvqBpYPnQ", 6),
        ("_KVGPoKJwEeusg-blS2y7tg", 18),
        ("_ZN1JAKJxEeusg-blS2y7tg", 21),
        ("_MDzzEAWrEeyIAukYgdUX5A", 11),
        ("_VKTjIIQTEeyxv9w6U6_UQg", 26),
        ("_dr0dQLB8EeqLtsTDMgEHBw", 48),
        ("_gRrk4bB_EeqLtsTDMgEHBw", 11),
        ("__dKYsdZLEeqiU8uzTY0Puw", 21),
        ("_jGZ6geCUEeu-FsE4WGMncQ", 4),
        ("_NfMDoAATEeykFulkDFvkrg", 8),
    ],
)
def test_iter_visible(model, diag_uid, num_nodes):
    diag = model.diagrams.by_uuid(diag_uid)
    diagram_elements = sum(
        1 for _ in aird.iter_visible(model._loader, diag._element)
    )

    assert diagram_elements == num_nodes



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Model-level PVMT configuration."""

from __future__ import annotations

__all__ = [
    "NS",
    "ManagedDomain",
    "ManagedGroup",
    "PVMTConfiguration",
    "PVMTDescriptionProperty",
    "ScopeError",
    "SelectorRules",
]

import collections.abc as cabc
import dataclasses
import logging
import operator
import re
import typing as t

import markupsafe
import typing_extensions as te
from lxml import etree

import capellambse
import capellambse.metamodel as mm
import capellambse.model as m
from capellambse import helpers

LOGGER = logging.getLogger(__name__)

PVMT_SCHEMA_VERSION = "1.0.0"

_PROP_OPS: dict[str, cabc.Callable[[t.Any, t.Any], bool]] = {
    "=": operator.eq,
    "!": operator.ne,
    "?": operator.contains,
    "&": str.startswith,
    "#": str.endswith,
    "<": operator.lt,
    ">": operator.gt,
    "%": operator.le,
    ":": operator.ge,
}
_PROP_OPS_RE = "|".join(re.escape(op) for op in _PROP_OPS)
_PROP_RE = re.compile(
    rf"^(?P<prop>[^.]+\.[^.]+\.[^.]+)(?P<op>{_PROP_OPS_RE})(?P<val>.*)"
)
_RULES_RE = re.compile(
    r"(?m)\[(?P<key>\w+)\]\s*(?P<value>.+?)\s*\[/(?P=key)\]"
)

NS = m.Namespace(
    m.VIRTUAL_NAMESPACE_PREFIX + "pvmt",
    "capellambse.virtual.pvmt",
)


class ScopeError(m.InvalidModificationError):
    """Raised when trying to apply a PVMT group to an out-of-scope element."""

    @property
    def group(self) -> ManagedGroup:
        return self.args[0]

    @property
    def obj(self) -> m.ModelObject:
        return self.args[1]

    def __str__(self) -> str:
        groupname = f"{self.group.parent.name}.{self.group.name}"
        objrepr = getattr(self.obj, "_short_repr_", lambda: repr(self.obj))()
        return f"PVMT group {groupname!r} does not apply to {objrepr}"


def _matchprops(
    obj: m.ModelObject,
    props: tuple[tuple[str, str, str], ...],
) -> bool:
    for prop, op, wanted in props:
        group, prop = prop.rsplit(".", 1)
        try:
            actual = obj.property_value_groups[group][prop]  # type: ignore
        except (AttributeError, KeyError):
            return False

        cmp = _PROP_OPS[op]
        if isinstance(actual, float | int | str):
            ismatch = cmp(actual, type(actual)(wanted))
        elif isinstance(actual, bool):
            ismatch = cmp(actual, wanted == "true")
        elif isinstance(actual, mm.capellacore.EnumerationPropertyLiteral):
            ismatch = cmp(actual.name, wanted)
        else:
            raise TypeError(
                f"Unhandled property value type: {type(actual).__name__}"
            )

        if not ismatch:
            return False

    return True


@dataclasses.dataclass(frozen=True)
class SelectorRules:
    raw: str

    _model: capellambse.MelodyModel | None = dataclasses.field(
        default=None, repr=False
    )

    @property
    def classes(self) -> tuple[type[m.ModelObject], ...]:
        if self.raw == "":
            return ()

        if self._model is None:
            raise TypeError(
                f"This {type(self).__name__} is not associated with a model"
            )

        classes = []
        for match in _RULES_RE.finditer(self.raw):
            if match.group("key") != "CLASS":
                continue
            for uri in match.group("value").split(","):
                _, clsname = uri.rsplit("/", 1)
                classes.append(self._model.resolve_class(clsname))

        return tuple(classes)

    @property
    def layers(self) -> tuple[type[mm.cs.BlockArchitecture], ...]:
        classes: list[type[mm.cs.BlockArchitecture]] = []
        for match in _RULES_RE.finditer(self.raw):
            if match.group("key") != "ARCHITECTURE":
                continue

            for arch in match.group("value").split(";"):
                if arch == "OPERATIONAL":
                    classes.append(mm.oa.OperationalAnalysis)
                elif arch == "SYSTEM":
                    classes.append(mm.sa.SystemAnalysis)
                elif arch == "LOGICAL":
                    classes.append(mm.la.LogicalArchitecture)
                elif arch == "PHYSICAL":
                    classes.append(mm.pa.PhysicalArchitecture)
                else:
                    LOGGER.debug("Unknown ARCHITECTURE, ignoring: %r", arch)

        return tuple(classes)

    @property
    def properties(self) -> tuple[tuple[str, str, str], ...]:
        propdefs = []
        for rule in _RULES_RE.finditer(self.raw):
            if rule.group("key") != "PROPERTY":
                continue

            value = rule.group("value")
            match = _PROP_RE.fullmatch(value)
            if not match:
                raise RuntimeError(f"Invalid property spec: {value!r}")
            prop, op, val = match.group("prop", "op", "val")
            assert isinstance(prop, str)
            assert isinstance(op, str)
            assert isinstance(val, str)
            propdefs.append((prop, op, val))
        return tuple(propdefs)


class PVMTDescriptionProperty(m.BasePOD[SelectorRules]):
    def __init__(self, attribute: str) -> None:
        super().__init__(attribute, default=SelectorRules(""), writable=True)
        self.__doc__ = "The element selector rules for this group."

    def _from_xml(self, obj: m.ModelElement, data: str, /) -> SelectorRules:
        return SelectorRules(data, obj._model)

    def _to_xml(
        self,
        obj: m.ModelElement,
        value: SelectorRules | str,
        /,
    ) -> str:
        del obj
        if isinstance(value, SelectorRules):
            value = value.raw
        elif not isinstance(value, str):
            raise TypeError(
                "Unsupported type, must be str or SelectorRules:"
                f" {type(value).__name__}"
            )

        # TODO implement validation
        return value


class ManagedGroup(mm.capellacore.PropertyValueGroup):
    """A managed group of property values."""

    _required_attrs = frozenset({"name"})

    selector = PVMTDescriptionProperty("description")

    @property
    def fullname(self) -> str:
        return f"{self.parent.name}.{self.name}"

    def applies_to(self, obj: m.ModelObject) -> bool:
        """Determine whether this group applies to a model element."""
        classes = self.selector.classes
        if classes and type(obj) not in classes:
            return False
        layers = self.selector.layers
        if layers and type(getattr(obj, "layer", None)) not in layers:
            return False

        if props := self.selector.properties:
            return _matchprops(obj, props)
        return True

    def find_applicable(self) -> m.ElementList:
        """Find all elements in the model that this group applies to."""
        objs = self._model.search(*self.selector.classes)
        if layers := self.selector.layers:
            objs = objs.filter(
                lambda i: isinstance(getattr(i, "layer", None), layers)
            )
        if props := self.selector.properties:
            objs = objs.filter(lambda i: _matchprops(i, props))
        return objs

    def apply(self, obj: m.ModelObject) -> mm.capellacore.PropertyValueGroup:
        """Apply this group to the model element, and return the applied group.

        If the group is not applicable to the passed element, a
        :class:ScopeError will be raised.

        If the group was already applied earlier, that group instance
        will be returned instead. This makes it safe to always use this
        method when working with PVMT-managed property value groups.
        """
        if not self.applies_to(obj):
            raise ScopeError(self, obj)

        objrepr = getattr(obj, "_short_repr_", lambda: repr(obj))
        if not hasattr(obj, "property_value_groups"):
            raise TypeError(f"Object cannot own PV groups: {objrepr()}")
        if not hasattr(obj, "applied_property_value_groups"):
            raise TypeError(f"Cannot apply PV groups to {objrepr()}")

        groupname = f"{self.parent.name}.{self.name}"
        try:
            return obj.property_value_groups.by_name(groupname, single=True)
        except KeyError:
            pass

        groupobj = obj.property_value_groups.create(
            name=groupname,
            applied_property_value_groups=[
                m.wrap_xml(self._model, self._element)
            ],
        )
        for propdef in self.property_values:
            pv = groupobj.property_values.create(
                type(propdef).__name__,
                name=propdef.name,
                applied_property_values=[propdef],
                value=propdef.value,
            )
            if isinstance(propdef, mm.capellacore.EnumerationPropertyValue):
                pv.type = propdef.type
        assert hasattr(obj, "applied_property_value_groups")
        obj.applied_property_value_groups.append(groupobj)
        return groupobj

    def _short_html_(self) -> markupsafe.Markup:
        return helpers.make_short_html(
            type(self).__name__,
            self.uuid,
            self.fullname,
        )


class ManagedDomain(mm.capellacore.PropertyValuePkg):
    """A "domain" in the property value management extension."""

    _required_attrs = frozenset({"name"})

    version = property(
        lambda self: self.property_values.by_name("version").value
    )
    types = m.Containment[mm.capellacore.EnumerationPropertyType](
        "ownedEnumerationPropertyTypes",
        (mm.capellacore.NS, "EnumerationPropertyType"),
    )
    groups: m.Containment[mm.capellacore.PropertyValueGroup] = m.Containment(  # type: ignore[assignment]
        "ownedPropertyValueGroups",
        (mm.capellacore.NS, "PropertyValueGroup"),
        mapkey="name",
        alternate=ManagedGroup,
    )

    def __init__(
        self,
        model: capellambse.MelodyModel,
        parent: etree._Element,
        xmltag: str | None = None,
        /,
        **kw: t.Any,
    ) -> None:
        super().__init__(model, parent, xmltag, **kw)
        self.property_values.create(name="version", value=PVMT_SCHEMA_VERSION)

    @classmethod
    def from_model(
        cls, model: capellambse.MelodyModel, element: etree._Element
    ) -> te.Self:
        self = m.wrap_xml(model, element, cls)
        try:
            version = self.property_values.by_name("version").value
        except Exception:
            self.property_values.create(
                name="version", value=PVMT_SCHEMA_VERSION
            )
        else:
            if version != PVMT_SCHEMA_VERSION:
                raise RuntimeError(
                    f"Unsupported schema version {version!r}"
                    f" on PVMT element {self._short_repr_()}"
                )
        return self


class PVMTConfiguration(mm.capellacore.PropertyValuePkg):
    """Provides access to the model-wide PVMT configuration."""

    domains = m.Containment[mm.capellacore.PropertyValuePkg](
        "ownedPropertyValuePkgs",
        (mm.capellacore.NS, "PropertyValuePkg"),
        mapkey="name",
        alternate=ManagedDomain,
    )



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import capellambse
import capellambse.metamodel as mm
import capellambse.model as m

from . import _validate
from ._validate import rule, virtual_type


@virtual_type(mm.sa.Capability)
def SystemCapability(_: mm.sa.Capability) -> bool:
    return True


@virtual_type(mm.sa.SystemComponent)
def SystemActor(cmp: mm.sa.SystemComponent) -> bool:
    return cmp.is_actor


@virtual_type(mm.sa.SystemComponent)
def SystemComponent(cmp: mm.sa.SystemComponent) -> bool:
    return not cmp.is_actor


@virtual_type(mm.pa.PhysicalComponent)
def BehaviourPhysicalComponent(cmp: mm.pa.PhysicalComponent) -> bool:
    return cmp.nature == mm.modeltypes.PhysicalComponentNature.BEHAVIOR


@virtual_type(mm.capellacommon.State)
def OperationalState(state: mm.capellacommon.State) -> bool:
    return isinstance(state.layer, mm.oa.OperationalAnalysis)


@virtual_type(mm.capellacommon.State)
def SystemState(state: mm.capellacommon.State) -> bool:
    return isinstance(state.layer, mm.sa.SystemAnalysis)


@virtual_type(mm.capellacommon.State)
def LogicalState(state: mm.capellacommon.State) -> bool:
    return isinstance(state.layer, mm.la.LogicalArchitecture)


@virtual_type(mm.capellacommon.State)
def PhysicalState(state: mm.capellacommon.State) -> bool:
    return isinstance(state.layer, mm.pa.PhysicalArchitecture)


@virtual_type(mm.fa.FunctionalExchange)
def PhysicalFunctionExchange(fex: mm.fa.FunctionalExchange) -> bool:
    return isinstance(fex.layer, mm.pa.PhysicalArchitecture)


# 00. Common
@rule(
    category=_validate.Category.RECOMMENDED,
    types=[
        mm.sa.Capability,
        mm.sa.SystemFunction,
        SystemActor,
        SystemComponent,
        mm.oa.Entity,
        mm.oa.OperationalCapability,
        mm.oa.OperationalActivity,
        mm.la.LogicalFunction,
        mm.la.LogicalComponent,
        mm.pa.PhysicalFunction,
        mm.pa.PhysicalComponent,
        mm.capellacommon.State,
    ],
    id="Rule-001",
    name="Object has a description or summary",
    rationale=(
        "A comprehensive description or summary for an object is essential to"
        " ensure a clear understanding of its purpose, function, and role"
        " within the system. Providing a concise, yet informative description"
        " fosters better collaboration among team members, reduces ambiguity,"
        " and facilitates efficient decision-making."
    ),
    action="fill the description and/or summary text fields",
)
def has_non_empty_description_or_summary(
    obj: capellambse.model.ModelElement,
) -> bool:
    return bool(obj.description) or bool(obj.summary)


@rule(
    category=_validate.Category.REQUIRED,
    types=[
        mm.sa.Capability,
        mm.oa.OperationalCapability,
    ],
    id="Rule-002",
    name="Capability involves an Entity / Actor",
    rationale=(
        "Each Capability serves a need and brings a benefit for at least one"
        " of the Actors/Entities. By involving an actor / entity in a"
        " Capability we explicitly name stakeholders behind the Capability."
    ),
    action=(
        "Add at least one involved Actor or Entity,"
        " or include the Capability in another one."
    ),
)
def capability_involves_entity(obj: capellambse.model.ModelElement) -> bool:
    if isinstance(obj, mm.oa.OperationalCapability):
        has_involvements = bool(obj.involved_entities)
    else:
        assert isinstance(obj, mm.sa.Capability)
        has_involvements = bool(obj.involved_components)
    return has_involvements or bool(obj.included_by)


@rule(
    category=_validate.Category.RECOMMENDED,
    types=[
        mm.sa.Capability,
        mm.oa.OperationalCapability,
        mm.la.CapabilityRealization,
        # The CapabilityRealization in the pa layer is also affected by this
        # rule but is not defined as pa.CapabilityRealization since its the
        # same class as la.CapabilityRealization
    ],
    id="Rule-004",
    name="Capability has a defined pre-condition",
    rationale=(
        "Defining a pre-condition for a Capability helps clarify the necessary"
        " state of the primary Actor before interacting with the System, which"
        " enables a better understanding of the context and ensures"
        " prerequisites are met for successful system performance within the"
        " scope of the Capability."
    ),
    action=(
        "Define a pre-condition for this Capability that describes the initial"
        " state or requirements of the primary Actor before interacting with"
        " the System, to provide clarity on the starting context for the"
        " Capability."
    ),
)
def has_precondition(obj: capellambse.model.ModelElement) -> bool:
    return obj.precondition is not None


@rule(
    category=_validate.Category.RECOMMENDED,
    types=[
        mm.sa.Capability,
        mm.oa.OperationalCapability,
        mm.la.CapabilityRealization,
    ],
    id="Rule-005",
    name="Capability has a defined post-condition",
    rationale=(
        "Defining a post-condition for a Capability helps establish the"
        " expected state of the primary Actor after interacting with the"
        " System, ensuring clear understanding of the desired outcome and"
        " enabling effective evaluation of system performance within the scope"
        " of the Capability."
    ),
    action=(
        "Define a post-condition for this Capability that describes the"
        " expected state or outcome for the primary Actor after interacting"
        " with the System, to ensure clear understanding of the desired"
        " results and enable effective evaluation of system performance."
    ),
)
def has_postcondition(obj):
    return obj.postcondition is not None


@rule(
    category=_validate.Category.REQUIRED,
    types=[PhysicalFunctionExchange],
    id="Rule-006",
    name=(
        "All Functional exchanges shall be allocated to at least one component"
        " exchange"
    ),
    rationale=(
        "Each functional exchange should be allocated to any one of the"
        " component exchange. Otherwise this implies that the information"
        " flow is not considered or apportioned in the architecture."
    ),
    action=(
        "Allocated the functional exchange to the appropriate component"
        " exchange"
    ),
)
def functional_exchange_allocated_to_component_exchange(
    obj: m.ModelElement,
) -> bool:
    return bool(obj.allocating_component_exchange)


@rule(
    category=_validate.Category.REQUIRED,
    types=[mm.sa.SystemFunction, mm.oa.OperationalActivity],
    id="Rule-011",
    name="A Behavior element shall be allocated to a structure element",
    rationale=(
        "An unallocated Behavior element implies that it is useless functionality"
        " or the modelling is incomplete."
    ),
    action="Allocate a Behavior Element to a Structure Element",
)
def behavior_element_allocated_to_structure_element(
    obj: m.ModelElement,
) -> bool:
    assert isinstance(obj, mm.sa.SystemFunction | mm.oa.OperationalActivity)
    return obj.owner is not None


# 01. Operational Analysis
@rule(
    category=_validate.Category.REQUIRED,
    types=[mm.oa.OperationalCapability],
    id="Rule-007",
    name=(
        "Capability involves at least two Activities from different"
        " entities/actors"
    ),
    rationale=(
        "An operational capability should have at least two operational"
        " activities in a sequence, performed by two different entity/actor to"
        " qualify it to be meaningful."
    ),
    action=(
        "involve two operational activities from two different entity/actor"
    ),
)
def capability_involves_two_activities_from_different_entities(
    obj: m.ModelElement,
) -> bool:
    actors = {
        activity.owner.uuid
        for activity in obj.involved_activities
        if activity.owner is not None
    }
    return len(actors) > 1


@rule(
    category=_validate.Category.REQUIRED,
    types=[mm.oa.OperationalActivity],
    id="Rule-008",
    name="Activity has at least one interaction with another activity",
    rationale=(
        "To have any effect in the context of the model, an Operational "
        " Activity needs to interact (i.e. receive inputs from and/or send "
        " outputs to) with the rest of the model, i.e. with other Operational "
        " Activities."
    ),
    action=(
        "Define an interaction for the operational activity with another"
        " activity"
    ),
)
def activity_has_interaction_with_another_activity(
    obj: m.ModelElement,
) -> bool:
    return bool(obj.related_exchanges)


# 02. System Analysis
@rule(
    category=_validate.Category.REQUIRED,
    types=[mm.sa.Capability],
    id="Rule-009",
    name=(
        "System capability involves at least one Actor Function and one"
        " System function"
    ),
    rationale=(
        "A well-defined Capability should involve at least one Actor Function"
        " and one system function. This helps understanding the functional"
        " contribution of an Actor in the scope of the Capability."
    ),
    action=(
        "involve at least one actor function and system function in the"
        " Capability"
    ),
)
def capability_involves_actor_and_system_function(
    obj: m.ModelElement,
) -> bool:
    actor_functions = [
        fnc
        for fnc in obj.involved_functions
        if fnc.owner and fnc.owner.is_actor
    ]
    system_functions = [
        fnc
        for fnc in obj.involved_functions
        if fnc.owner and not fnc.owner.is_actor
    ]
    return len(actor_functions) > 0 and len(system_functions) > 0


@rule(
    category=_validate.Category.REQUIRED,
    types=[mm.sa.Capability],
    id="Rule-010",
    name="IS- and SHOULD entity involvements match",
    rationale=(
        "Capability should involve both Actors and Functions, with Functions"
        " allocated to respective entities (Actor or System). A mismatch"
        " between these two lists might suggest that a function is linked to"
        " an uninvolved actor, or an actor is involved without having any"
        " associated functions for that capability. Ensuring proper alignment"
        " helps maintain logical consistency and efficiency in the system"
        " design."
    ),
    action=(
        "To correct the mismatch, review and update the Capability's involved"
        " Actors and Functions, ensuring each Actor contributes at least one"
        " Function, and each Function is allocated to an appropriate Actor or"
        " System"
    ),
)
def is_and_should_entity_involvements_match(obj: m.ModelElement) -> bool:
    is_involvements = {x.owner.uuid for x in obj.involved_functions if x.owner}
    should_involvements = {x.uuid for x in obj.involved_components}
    return is_involvements == should_involvements


@rule(
    category=_validate.Category.RECOMMENDED,
    types=[mm.sa.SystemFunction],
    id="SF-040",
    name="Function shall have at least one input or output",
    rationale=(
        "A Function without inputs or outputs may not effectively contribute"
        " to the overall model, as it would not interact with other functions."
    ),
    action="consider adding inputs and / or outputs to the Function.",
)
def function_has_inputs_and_outputs(obj: m.ModelElement) -> bool:
    return len(obj.inputs) > 0 or len(obj.outputs) > 0



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

import collections.abc as cabc
import contextlib
import logging
import pathlib
import shutil

import click

import capellambse
import capellambse.model as m
from capellambse import _native, cli_helpers, helpers
from capellambse.extensions import filtering

_LOGGER = logging.getLogger(__name__)


@click.command()
@click.option(
    "-m",
    "--model",
    "model_",
    type=cli_helpers.ModelCLI(),
    required=True,
    help="Source Capella model to derive from",
)
@click.option(
    "-r",
    "--result",
    "result_strings",
    multiple=True,
    help="A filtering result object, specified either by UUID or name.",
)
@click.option(
    "-o",
    "--output",
    type=click.Path(file_okay=False, dir_okay=True, path_type=pathlib.Path),
    default=".",
    help="Directory to store the derived project(s) in.",
)
@click.option(
    "-p",
    "--name-pattern",
    default="{model.name}_derived_{result.name}",
    help=(
        "The 'str.format()' pattern for generating output project names."
        " Available format arguments are 'model' and the current 'result'."
    ),
)
@click.option("--exe", help="Name or path of the Capella executable")
@click.option("--docker", help="Name of a Docker image containing Capella")
def main(
    model_: capellambse.MelodyModel,
    result_strings: list[str],
    output: pathlib.Path,
    name_pattern: str,
    exe: str | None,
    docker: str | None,
) -> None:
    """Manage filtered Capella projects.

    Currently the only supported COMMAND is "derive", which performs
    project derivation based on an existing filtering result.

    The native Capella installation (options ``--exe`` or ``--docker``)
    may contain the placeholder '{VERSION}', which will be replaced by
    the Capella version number that the model was created with (e.g.
    "6.0.0"). If you have docker installed, a good default value is:

    \b
    --docker ghcr.io/dbinfrago/capella-dockerimages/capella/base:{VERSION}-selected-dropins-main

    Exactly one of ``--exe`` or ``--docker`` must be specified.
    """  # noqa: D301
    if exe and docker:
        raise click.UsageError("--exe and --docker are mutually exclusive")
    if not exe and not docker:
        raise click.UsageError("Exactly one of --exe or --docker is required")

    results = _find_results(model_, result_strings)
    output.mkdir(exist_ok=True, parents=True)
    with (
        _native.native_capella(model_, exe=exe, docker=docker) as cli,
        click.progressbar(
            results, item_show_func=lambda i: getattr(i, "name", "")
        ) as progress,
    ):
        for result in progress:
            derived_name = name_pattern.format(model=model_, result=result)
            output_sub = output / derived_name
            with contextlib.suppress(FileNotFoundError):
                shutil.rmtree(output_sub)
            proc = cli(
                *_native.ARGS_CMDLINE,
                "-appid",
                "org.polarsys.capella.filtering.commandline",
                "-input",
                "main_model",
                "-filteringresultid",
                result.uuid,
                "-derivationprojectname",
                derived_name,
            )
            derived = cli.workspace / derived_name
            if not derived.exists():
                _LOGGER.error(
                    "Failed to derive result %r (%s)\n\n%s",
                    result.name,
                    result.uuid,
                    proc.stdout,
                )
                raise RuntimeError(
                    "Output directory does not exist for derived result"
                    f" {result.name!r} ({result.uuid})"
                )
            shutil.copytree(derived, output_sub)


def _find_results(
    loaded_model: capellambse.MelodyModel, result_strings: list[str]
) -> cabc.Sequence[filtering.AbstractFilteringResult]:
    all_results: m.ElementList[filtering.AbstractFilteringResult]
    all_results = loaded_model.search(filtering.AbstractFilteringResult)
    if not result_strings:
        return all_results

    wanted: list[filtering.AbstractFilteringResult] = []
    obj: m.ModelElement
    for result in result_strings:
        if helpers.is_uuid_string(result):
            try:
                obj = loaded_model.by_uuid(result)
            except KeyError:
                pass
            else:
                if not isinstance(obj, filtering.AbstractFilteringResult):
                    click.echo(
                        f"Error: Object {obj._short_repr_()} is of type"
                        f" {type(obj).__name__}, expected"
                        f" an AbstractFilteringResult",
                        err=True,
                    )
                    raise SystemExit(1)
                wanted.append(obj)
                continue

        obj = all_results.by_name(result, single=True)
        assert isinstance(obj, filtering.AbstractFilteringResult)
        wanted.append(obj)
    return wanted



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import collections.abc as cabc
import errno
import itertools
import logging
import os
import pathlib
import re
import typing as t
import urllib.parse

import requests
import typing_extensions as te

from capellambse import helpers

from . import abc

LOGGER = logging.getLogger(__name__)

NOT_FOUND_CODES = (
    400,  # Bad Request
    404,  # Not Found
    410,  # Gone
    418,  # I'm a teapot
)


class DownloadStream(t.BinaryIO):
    __stream: cabc.Iterator[bytes]
    __buffer: memoryview

    def __init__(
        self, session: requests.Session, url: str, chunk_size: int = 1024**2
    ) -> None:
        LOGGER.debug("Opening HTTP download stream from %s", url)
        self.url = url
        self.chunk_size = chunk_size

        response = session.get(self.url, stream=True)
        LOGGER.debug("Status: %d %s", response.status_code, response.reason)
        if response.status_code in NOT_FOUND_CODES:
            raise FileNotFoundError(
                errno.ENOENT,
                f"Got {response.status_code} {response.reason} for URL {url}",
            )
        response.raise_for_status()
        self.__stream = response.iter_content(
            self.chunk_size, decode_unicode=False
        )
        self.__buffer = memoryview(b"")

    def __enter__(self) -> DownloadStream:
        return self

    def __exit__(self, *args: t.Any) -> None:
        self.close()

    def read(self, n: int = -1) -> bytes:
        if n == -1:
            return b"".join(itertools.chain((self.__buffer,), self.__stream))

        if not self.__buffer:
            try:
                self.__buffer = memoryview(next(self.__stream))
            except StopIteration:
                return b""

        chunk = bytes(self.__buffer[:n])
        self.__buffer = self.__buffer[n:]
        return chunk

    def readable(self) -> bool:
        return True

    def write(self, s: bytes | bytearray) -> int:  # type: ignore[override]
        del s
        raise TypeError("Cannot write to a read-only stream")

    def writable(self) -> bool:
        return False

    def close(self) -> None:
        del self.__stream
        del self.__buffer


class HTTPFileHandler(abc.FileHandler):
    """A remote file handler that fetches files using HTTP GET."""

    def __init__(
        self,
        path: str | os.PathLike,
        username: str | None = None,
        password: str | None = None,
        *,
        headers: cabc.Mapping[str, str] | None = None,
        subdir: str | pathlib.PurePosixPath = "/",
    ) -> None:
        """Connect to a remote server through HTTP or HTTPS.

        This file handler supports two ways of specifying a URL:

        1. If a plain URL is passed, the requested file name is appended
           after a forward slash ``/``.
        2. The URL may contain one or more of the following escape
           sequences to provide more fine-grained control over how and
           where the file name is inserted into the URL:

           - ``%s``: The path to the file, with everything except
             forward slashes percent-escaped
           - ``%q``: The path to the file, with forward slashes percent
             escaped as well
           - ``%d``: The directory name, without trailing slash, like %s
           - ``%n``: The file name without extension
           - ``%e``: The file extension without leading dot
           - ``%%``: A literal percent sign

        Examples: When requesting the file name ``demo/my model.aird``,
        ...

        - ``https://example.com/~user`` as ``path`` results in the URL
          ``https://example.com/~user/demo/my%20model.aird``
        - ``https://example.com/~user/%s`` results in
          ``https://example.com/~user/demo/my%20model.aird``
        - ``https://example.com/?file=%q`` results in
          ``https://example.com/?file=demo%2Fmy%20model.aird``

        Note that the file name that is inserted into the URL will never
        start with a forward slash. This means that a URL like
        ``https://example.com%s`` will not work; you need to hard-code
        the slash at the appropriate place.

        This also applies to the ``%q`` escape. If the server expects
        the file name argument to start with a slash, hard-code a
        percent-escaped slash in the URL. For example, instead of
        ``...?file=%q`` use ``...?file=%2F%q``.

        Parameters
        ----------
        path
            The base URL to fetch files from. Must start with
            ``http://`` or ``https://``. See above for how to specify
            complex URLs.
        username
            The username for HTTP Basic Auth.
        password
            The password for HTTP Basic Auth.
        headers
            Additional HTTP headers to send to the server.
        subdir
            Prepend this path to all requested files. It is subject to
            the same file name escaping rules explained above.
        """
        if not isinstance(path, str):
            raise TypeError(
                "HTTPFileHandler requires a str path, not"
                f" {type(path).__name__}"
            )
        if bool(username) != bool(password):
            raise ValueError(
                "Either both username and password must be given, or neither"
            )

        if not re.search("%[%a-z]", path):
            path = path.rstrip("/") + "/%s"

        super().__init__(path, subdir=subdir)

        self.session = requests.Session()
        self.session.headers.update(headers or {})
        if username and password:
            self.session.auth = (username, password)

    def open(
        self,
        filename: str | pathlib.PurePosixPath,
        mode: t.Literal["r", "rb", "w", "wb"] = "rb",
    ) -> t.BinaryIO:
        if "w" in mode:
            raise NotImplementedError("Cannot upload to HTTP(S) locations")
        assert isinstance(self.path, str)
        fname = self.subdir / helpers.normalize_pure_path(filename)
        fname_str = str(fname).lstrip("/")
        q = urllib.parse.quote
        replace = {
            "%s": q(fname_str, safe="/"),
            "%q": q(fname_str, safe=""),
            "%d": q(str(fname.parent).lstrip("/")),
            "%n": q(fname.with_suffix("").name),
            "%e": q(fname.suffix.lstrip(".")),
            "%%": "%",
        }
        url = re.sub("%[%a-z]", lambda m: replace[m.group(0)], self.path)
        assert url != self.path
        return DownloadStream(  # type: ignore[abstract] # false-positive
            self.session, url
        )

    def write_transaction(self, **kw: t.Any) -> t.NoReturn:
        raise NotImplementedError(
            "Write transactions for HTTP(S) are not implemented"
        )

    def iterdir(  # pragma: no cover
        self, path: str | pathlib.PurePosixPath = ".", /
    ) -> cabc.Iterator[abc.FilePath[te.Self]]:
        del path
        raise TypeError(
            "Cannot list files on raw HTTP sources."
            " Maybe you forgot a 'git+' prefix?"
        )



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import capellambse.model as m

from . import modellingcore
from . import namespaces as ns

NS = ns.BEHAVIOR


class AbstractBehavior(modellingcore.AbstractNamedElement, abstract=True):
    """Abstract base class for behaviors."""

    is_control_operator = m.BoolPOD("isControlOperator")
    parameter_sets = m.Association["modellingcore.AbstractParameterSet"](
        (ns.MODELLINGCORE, "AbstractParameterSet"), "ownedParameterSet"
    )
    parameters = m.Association["modellingcore.AbstractParameter"](
        (ns.MODELLINGCORE, "AbstractParameter"), "ownedParameter"
    )


class AbstractSignal(modellingcore.AbstractType, abstract=True):
    """Abstract base class for signals."""


class AbstractEvent(modellingcore.AbstractType, abstract=True):
    """Specification of an occurrence that may trigger effects."""


class AbstractTimeEvent(AbstractEvent, abstract=True):
    """A point in time.

    A time event specifies a point in time by an expression. The
    expression might be absolute or might be relative to some other
    point in time.
    """

    is_relative = m.BoolPOD("isRelative")
    when = m.Single["TimeExpression"](
        m.Association((NS, "TimeExpression"), "when")
    )


class AbstractMessageEvent(AbstractEvent, abstract=True):
    """The receipt by an object of either a call or a signal."""


class AbstractSignalEvent(AbstractMessageEvent, abstract=True):
    """The receipt of an asynchronous signal.

    A signal event may cause a response, such as a state machine
    transition as specified in the classifier behavior of the classifier
    that specified the receiver object, if the signal referenced by the
    send request is mentioned in a reception owned or inherited by the
    classifier that specified the receiver object.
    """

    signal = m.Single["AbstractSignal"](
        m.Association((NS, "AbstractSignal"), "signal")
    )


class TimeExpression(modellingcore.ValueSpecification, abstract=True):
    """A specification of a point in time."""

    observations = m.Association["modellingcore.AbstractNamedElement"](
        (ns.MODELLINGCORE, "AbstractNamedElement"), "observations"
    )
    expression = m.Single["modellingcore.ValueSpecification"](
        m.Association((ns.MODELLINGCORE, "ValueSpecification"), "expression")
    )



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Enumeration types used by the MelodyModel."""

import warnings

warnings.warn(
    (
        "The 'modeltypes' module is deprecated,"
        " use the enums from their respective source module instead"
    ),
    DeprecationWarning,
    stacklevel=2,
)

from .activity import ObjectNodeKind as ObjectNodeKind
from .activity import ObjectNodeOrderingKind as ObjectNodeOrderingKind
from .capellacommon import ChangeEventKind as ChangeEventKind
from .capellacommon import TimeEventKind as TimeEventKind
from .capellacommon import TransitionKind as TransitionKind
from .capellacore import VisibilityKind as VisibilityKind
from .epbs import ConfigurationItemKind as ConfigurationItemKind
from .fa import ComponentExchangeKind as ComponentExchangeKind
from .fa import ComponentPortKind as ComponentPortKind
from .fa import ControlNodeKind as ControlNodeKind
from .fa import FunctionalChainKind as FunctionalChainKind
from .fa import FunctionKind as FunctionKind
from .fa import OrientationPortKind as OrientationPortKind
from .information import AggregationKind as AggregationKind
from .information import CollectionKind as CollectionKind
from .information import ElementKind as ElementKind
from .information import ExchangeMechanism as ExchangeMechanism
from .information import ParameterDirection as ParameterDirection
from .information import PassingMode as PassingMode
from .information import SynchronismKind as SynchronismKind
from .information import UnionKind as UnionKind
from .information.communication import (
    CommunicationLinkKind as CommunicationLinkKind,
)
from .information.communication import (
    CommunicationLinkProtocol as CommunicationLinkProtocol,
)
from .information.datatype import NumericTypeKind as NumericTypeKind
from .information.datavalue import BinaryOperator as BinaryOperator
from .information.datavalue import UnaryOperator as UnaryOperator
from .interaction import InteractionOperatorKind as InteractionOperatorKind
from .interaction import MessageKind as MessageKind
from .interaction import ScenarioKind as ScenarioKind
from .libraries import AccessPolicy as AccessPolicy
from .modellingcore import ParameterEffectKind as ParameterEffectKind
from .modellingcore import RateKind as RateKind
from .pa import PhysicalComponentKind as PhysicalComponentKind
from .pa import PhysicalComponentNature as PhysicalComponentNature
from .re import CatalogElementKind as CatalogElementKind
